from typing import Generator, Any, Union, Literal
from delta.tables import DeltaTable
from pyspark.sql import Column, DataFrame
import pyspark.sql.functions as fn
from pyspark.sql.window import Window

import trend_etl.schema.fields as fields
from trend_etl.schema.types import Merge_Strategy
import trend_etl.schema.args as arguments


def get_silver_table(_catalog: str, table_index: int) -> str:
    return arguments.tables[table_index]


# arguments:
#   target_catalog: name of catalog for bronze table
#   target_schema: name of schema for bronze table
#   target_table: name of table for bronze table
def full_table_name(target_catalog: str, target_schema: str, target_table: str) -> str:
    return f"{target_catalog}.{target_schema}.{target_table}"


def null_if_deleted(col: Column) -> Column:
    return fn.when((col == "[removed]") | (col == "[deleted]"), None).otherwise(col)


def partition_with_remaining(
    lst: list[str], *selections: list[str]
) -> Generator[list[str], Any, None]:
    for selection in selections:
        partition = []
        for x in selection:
            lst.remove(x)
            partition.append(x)
        yield partition
    assert len(lst) > 0
    yield lst


# NOTE: I could not test this is pyspark without getting a memory error.
#       I tested in a temporary notebook on databricks.
def get_canonical_reddit(
    input_df: DataFrame, schema: Literal["submission", "reply"]
) -> DataFrame:
    most_recent = most_recent_version(input_df, fn.col("source_keys.name"))
    earliest = earliest_version(input_df, fn.col("source_keys.name"))
    earliest_text = earliest_version(
        input_df.where(fn.col("text_version").isNotNull()),
        fn.col("source_keys.name"),
    )

    def favor_earliest(field: str, alias: str) -> Column:
        return fn.coalesce(
            fn.col(f"earliest.{field}"),
            fn.col(f"most_recent.{field}"),
        ).alias(alias)

    def use_most_recent(field: str, alias: str) -> Column:
        return fn.col(f"most_recent.{field}").alias(alias)

    def favor_earliest_text(field: str, alias: str) -> Column:
        return fn.coalesce(
            fn.col(f"earliest_text.{field}"),
            fn.col(f"earliest.{field}"),
        ).alias(alias)

    def merge_author() -> Column:
        return fn.coalesce(
            null_if_deleted(fn.col("earliest.source_keys.author")),
            null_if_deleted(fn.col("most_recent.source_keys.author")),
            fn.col("earliest.source_keys.author"),
            fn.col("most_recent.source_keys.author"),
        ).alias("author")

    def source_keys() -> Column:
        if schema == "submission":
            return fn.struct(
                favor_earliest("source_keys.name", "name"),
                merge_author(),
                favor_earliest("source_keys.author_fullname", "author_fullname"),
                favor_earliest("source_keys.subreddit_id", "subreddit_id"),
            ).alias("source_keys")
        else:  # reply
            return fn.struct(
                favor_earliest("source_keys.name", "name"),
                favor_earliest("source_keys.submission_id", "submission_id"),
                favor_earliest("source_keys.parent_id", "parent_id"),
                merge_author(),
                favor_earliest("source_keys.author_fullname", "author_fullname"),
                favor_earliest("source_keys.subreddit_id", "subreddit_id"),
            ).alias("source_keys")

    projection = [
        (
            use_most_recent(col, col)
            if col
            in (
                "num_comments",
                "num_crossposts",
                "score",
                "ups",
                "downs",
                "upvote_ratio",
            )
            else (
                source_keys()
                if col == "source_keys"
                else (
                    favor_earliest_text(col, col)
                    if col in ("text_version")
                    else favor_earliest(col, col)
                )
            )
        )
        for col in input_df.columns
        if col not in ("source_schema",)  # ignore these fields
    ] + [use_most_recent("ingested_on", "scores_collected_on")]

    canonical = (
        earliest.alias("earliest")
        .join(
            most_recent.alias("most_recent"),
            fn.col("most_recent.source_keys.name")
            == fn.col("earliest.source_keys.name"),
            "left_outer",
        )
        .join(
            earliest_text.alias("earliest_text"),
            fn.col("earliest_text.source_keys.name")
            == fn.col("earliest.source_keys.name"),
            "left_outer",
        )
        .select(*projection)
    )

    return canonical


# arguments:
#   input_df: the dataframe in bronze schema
#   table: full name of bronze table
def append_delta_table(input_df: DataFrame, table: str) -> None:
    (input_df.write.format("delta").mode("append").saveAsTable(table))


# Postcondition:
# Returned dataframe with earliest collected on with a
# "tie-breaker" going to the one with non-deleted text
def reddit_docs_select_earliest(input_df: DataFrame) -> DataFrame:
    collected_order = fn.col(fields.bronze_timestamp_field).asc()
    text_priority_order = (
        fn.when((fn.col("text") == "[removed]") | (fn.col("text") == "[deleted]"), 1)
        .otherwise(0)
        .asc()
    )

    window_spec = Window.partitionBy(fields.idents_doc_id_field).orderBy(
        text_priority_order, collected_order
    )

    with_row_num = input_df.withColumn("row_num", fn.row_number().over(window_spec))

    return with_row_num.where(fn.col("row_num") == 1).drop("row_num")


def select_earliest_by_key(input_df: DataFrame, keys: list[str]) -> DataFrame:
    collected_order = fn.col(fields.bronze_timestamp_field).asc()
    window_spec = Window.partitionBy(keys).orderBy(collected_order)

    with_row_num = input_df.withColumn("row_num", fn.row_number().over(window_spec))

    return with_row_num.where(fn.col("row_num") == 1).drop("row_num")


def select_recent_by_key(input_df: DataFrame, keys: list[str]) -> DataFrame:
    collected_order = fn.col(fields.bronze_timestamp_field).desc()
    window_spec = Window.partitionBy(keys).orderBy(collected_order)

    with_row_num = input_df.withColumn("row_num", fn.row_number().over(window_spec))

    return with_row_num.where(fn.col("row_num") == 1).drop("row_num")


def ao_merge_using_ingestion_key(delta_table: DeltaTable, batch: DataFrame) -> None:
    updates = batch.drop_duplicates([fields.ingest_version_field])

    merge_cond = (
        f"updates.{fields.ingest_version_field} = target.{fields.ingest_version_field}"
    )

    delta_table.alias("target").merge(
        updates.alias("updates"),
        merge_cond,
    ).whenNotMatchedInsertAll().execute()


def ao_merge_using_text_key(delta_table: DeltaTable, batch: DataFrame) -> None:
    updates = batch.drop_duplicates([fields.text_version_field])

    merge_cond = (
        f"updates.{fields.text_version_field} = target.{fields.text_version_field}"
    )

    delta_table.alias("target").merge(
        updates.alias("updates"),
        merge_cond,
    ).whenNotMatchedInsertAll().execute()


# Postcondition:
# Use earliest collected on with a "tie-breaker" going to the one with non-deleted text
def merge_earliest_collected_reddit_docs(
    delta_table: DeltaTable, micro_batch: DataFrame
) -> None:
    # NOTE: within each `micro_batch` choose the earliest collected
    updates = reddit_docs_select_earliest(micro_batch)

    merge_cond = (
        f"updates.{fields.idents_group_id_field} = target.{fields.idents_group_id_field}"
        f" and updates.{fields.idents_doc_id_field} = target.{fields.idents_doc_id_field}"
    )
    update_cond = (
        f"updates.{fields.bronze_timestamp_field} < target.{fields.bronze_timestamp_field}"
        " or ("
        f"updates.{fields.bronze_timestamp_field} <= target.{fields.bronze_timestamp_field}"
        " and updates.text != '[removed]'"
        " and updates.text != '[deleted]'"
        " and (target.text = '[removed]' or target.text = '[deleted]')"
        ")"
    )

    delta_table.alias("target").merge(
        updates.alias("updates"),
        merge_cond,
    ).whenNotMatchedInsertAll().whenMatchedUpdateAll(condition=update_cond).execute()


def most_recent_version(input_df: DataFrame, partition_col: Column) -> DataFrame:
    window_spec = Window.partitionBy("_partition").orderBy(
        fn.col("ingested_on").desc(), fn.col("ingest_version").desc()
    )

    return (
        input_df.withColumn("_partition", partition_col)
        .withColumn("_row_num", fn.row_number().over(window_spec))
        .where(fn.col("_row_num") == 1)
        .drop("_row_num", "_partition")
    )


def earliest_version(input_df: DataFrame, partition_col: Column) -> DataFrame:
    window_spec = Window.partitionBy("_partition").orderBy(
        fn.col("ingested_on").asc(), fn.col("ingest_version").asc()
    )

    return (
        input_df.withColumn("_partition", partition_col)
        .withColumn("_row_num", fn.row_number().over(window_spec))
        .where(fn.col("_row_num") == 1)
        .drop("_row_num", "_partition")
    )


def merge_earliest_collected_by_key(
    delta_table: DeltaTable, micro_batch: DataFrame, merge_keys: list[str]
) -> None:
    updates = select_earliest_by_key(micro_batch, merge_keys)

    merge_cond = " and ".join(
        map(lambda key: f"updates.{key} = target.{key}", merge_keys)
    )
    update_cond = f"updates.{fields.bronze_timestamp_field} < target.{fields.bronze_timestamp_field}"

    delta_table.alias("target").merge(
        updates.alias("updates"),
        merge_cond,
    ).whenNotMatchedInsertAll().whenMatchedUpdateAll(condition=update_cond).execute()


def merge_recent_collected_by_key(
    delta_table: DeltaTable, micro_batch: DataFrame, merge_keys: list[str]
) -> None:
    updates = select_recent_by_key(micro_batch, merge_keys)

    merge_cond = " and ".join(
        map(lambda key: f"updates.{key} = target.{key}", merge_keys)
    )
    update_cond = f"updates.{fields.bronze_timestamp_field} > target.{fields.bronze_timestamp_field}"

    delta_table.alias("target").merge(
        updates.alias("updates"),
        merge_cond,
    ).whenNotMatchedInsertAll().whenMatchedUpdateAll(condition=update_cond).execute()


def merge_drop_duplicates(
    input_df: DataFrame, strategy: Merge_Strategy, id_field: str, strategy_field: str
) -> DataFrame:
    merge_fn = fn.min if strategy == "keep-oldest" else fn.max
    picked = input_df.groupBy(id_field).agg(
        merge_fn(strategy_field).alias(strategy_field)
    )
    result = input_df.join(picked, [id_field, strategy_field], "left_semi")
    # NOTE: We cannot be sure each collection will not have duplicates itself.
    #       Therefore, we need to drop duplicates to ensure no duplicates.
    return result.drop_duplicates([id_field])


def merge_drop_duplicates_key(input_df: DataFrame, key: Column) -> DataFrame:
    key_field = "_key"

    return (
        input_df.withColumn(key_field, key).dropDuplicates([key_field]).drop(key_field)
    )


# Construct the most recent version from multiple silver versions of record.
def merge_most_recent_version(
    delta_table: DeltaTable,
    updates: DataFrame,
    merge_keys: list[str],
) -> None:
    merge_cond = " and ".join(
        map(lambda key: f"updates.{key} = target.{key}", merge_keys)
    )
    update_cond = (
        "updates.ingested_on > target.ingested_on"
        " or ("
        "updates.ingested_on = target.ingested_on"
        " AND updates.ingest_version > target.ingest_version"
        ")"
    )

    (
        delta_table.alias("target")
        .merge(updates.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .whenMatchedUpdateAll(update_cond)
        .execute()
    )


# Construct the earliest version from multiple silver versions of record.
def merge_earliest_version(
    delta_table: DeltaTable,
    updates: DataFrame,
    merge_keys: list[str],
) -> None:
    merge_cond = " and ".join(
        map(lambda key: f"updates.{key} = target.{key}", merge_keys)
    )
    update_cond = (
        "updates.ingested_on < target.ingested_on"
        " or ("
        "updates.ingested_on = target.ingested_on"
        " AND updates.ingest_version < target.ingest_version"
        ")"
    )

    (
        delta_table.alias("target")
        .merge(updates.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .whenMatchedUpdateAll(update_cond)
        .execute()
    )


# arguments
#   delta_table: delta table
# return: merge function (suitable for `foreachBatch`)
def uuid_upsert_to_delta(delta_table: DeltaTable, input_df: DataFrame) -> None:
    updates_df = input_df.dropDuplicates([fields.idents_source_id_field])

    (
        delta_table.alias("target")
        .merge(
            updates_df.alias("updates"),
            f"updates.{fields.idents_source_id_field} = target.{fields.idents_source_id_field}",
        )
        .whenNotMatchedInsertAll()
        .execute()
    )


# arguments
#   delta_table: delta table
#   strategy: merge strategy
# return: merge function (suitable for `foreachBatch`)
def facebook_posts_upsert_to_delta(
    delta_table: DeltaTable, strategy: Merge_Strategy, input_df: DataFrame
) -> None:
    updates_df = merge_drop_duplicates(
        input_df=input_df.withColumnRenamed(
            fields.bronze_timestamp_field, fields.first_collected_field
        ).withColumn(fields.last_collected_field, fn.expr("cast(null as timestamp)")),
        strategy=strategy,
        id_field=fields.idents_doc_id_field,
        strategy_field=fields.first_collected_field,
    )
    merge_cond = (
        f"updates.{fields.idents_group_id_field} = target.{fields.idents_group_id_field}"
        f" and updates.{fields.idents_doc_id_field} = target.{fields.idents_doc_id_field}"
    )
    update_cond_lt = f"updates.{fields.first_collected_field} < target.{fields.first_collected_field}"
    update_cond_gt = f"updates.{fields.first_collected_field} > target.{fields.first_collected_field}"

    (
        delta_table.alias("target")
        .merge(updates_df.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .whenMatchedUpdate(  # keep original title/text
            update_cond_lt,
            {
                fields.first_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
                fields.last_collected_field: fn.col(
                    f"target.{fields.first_collected_field}",
                ),
                fields.facebook_posts_title: fn.col(
                    f"updates.{fields.facebook_posts_title}",
                ),
                fields.facebook_posts_text: fn.col(
                    f"updates.{fields.facebook_posts_text}",
                ),
            },
        )
        .whenMatchedUpdate(  # update reactions to latest collected
            update_cond_gt,
            {
                fields.reactions_struct_field: fn.col(
                    f"updates.{fields.reactions_struct_field}",
                ),
                fields.facebook_posts_updated_at: fn.col(
                    f"updates.{fields.facebook_posts_updated_at}",
                ),
                fields.last_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
            },
        )
        .execute()
    )


# arguments
#   delta_table: delta table
#   strategy: merge strategy
# return: merge function (suitable for `foreachBatch`)
def facebook_comments_upsert_to_delta(
    delta_table: DeltaTable, strategy: Merge_Strategy, input_df: DataFrame
) -> None:
    updates_df = merge_drop_duplicates(
        input_df=input_df.withColumnRenamed(
            fields.bronze_timestamp_field, fields.first_collected_field
        ).withColumn(fields.last_collected_field, fn.expr("cast(null as timestamp)")),
        strategy=strategy,
        id_field=fields.idents_doc_id_field,
        strategy_field=fields.first_collected_field,
    )

    merge_cond = (
        f"updates.{fields.idents_group_id_field} = target.{fields.idents_group_id_field}"
        f" and updates.{fields.idents_doc_id_field} = target.{fields.idents_doc_id_field}"
    )
    update_cond_lt = f"updates.{fields.first_collected_field} < target.{fields.first_collected_field}"
    update_cond_gt = f"updates.{fields.first_collected_field} > target.{fields.first_collected_field}"

    (
        delta_table.alias("target")
        .merge(updates_df.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .whenMatchedUpdate(  # keep original text
            update_cond_lt,
            {
                fields.first_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
                fields.last_collected_field: fn.col(
                    f"target.{fields.first_collected_field}",
                ),
                fields.facebook_comments_text: fn.col(
                    f"updates.{fields.facebook_comments_text}",
                ),
            },
        )
        .whenMatchedUpdate(  # update reactions to latest collected
            update_cond_gt,
            {
                fields.reactions_struct_field: fn.col(
                    f"updates.{fields.reactions_struct_field}",
                ),
                fields.last_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
            },
        )
        .execute()
    )


# arguments
#   delta_table: delta table
#   strategy: merge strategy
# return: merge function (suitable for `foreachBatch`)
def facebook_groups_upsert_to_delta(
    delta_table: DeltaTable, strategy: Merge_Strategy, input_df: DataFrame
) -> None:
    updates_df = merge_drop_duplicates(
        input_df=input_df.withColumnRenamed(
            fields.bronze_timestamp_field, fields.first_collected_field
        ).withColumn(fields.last_collected_field, fn.expr("cast(null as timestamp)")),
        strategy=strategy,
        id_field=fields.idents_group_id_field,
        strategy_field=fields.first_collected_field,
    )

    merge_cond = f"updates.{fields.idents_group_id_field} = target.{fields.idents_group_id_field}"
    update_cond_lt = f"updates.{fields.first_collected_field} < target.{fields.first_collected_field}"
    update_cond_gt = f"updates.{fields.first_collected_field} > target.{fields.first_collected_field}"

    (
        delta_table.alias("target")
        .merge(updates_df.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .whenMatchedUpdate(
            update_cond_lt,
            {
                fields.first_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
                fields.last_collected_field: fn.col(
                    f"target.{fields.first_collected_field}",
                ),
            },
        )
        .whenMatchedUpdate(  # update name, description, member count
            update_cond_gt,
            {
                fields.last_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
                fields.facebook_groups_name: fn.col(
                    f"updates.{fields.facebook_groups_name}",
                ),
                fields.facebook_groups_description: fn.col(
                    f"updates.{fields.facebook_groups_description}",
                ),
                fields.facebook_groups_member_count: fn.col(
                    f"updates.{fields.facebook_groups_member_count}",
                ),
            },
        )
        .execute()
    )


# arguments
#   delta_table: delta table
#   strategy: merge strategy
# return: merge function (suitable for `foreachBatch`)
def subreddits_upsert_to_delta(
    delta_table: DeltaTable, strategy: Merge_Strategy, input_df: DataFrame
) -> None:
    updates_df = merge_drop_duplicates(
        input_df=input_df.withColumnRenamed(
            fields.bronze_timestamp_field, fields.first_collected_field
        ).withColumn(fields.last_collected_field, fn.expr("cast(null as timestamp)")),
        strategy=strategy,
        id_field=fields.idents_group_id_field,
        strategy_field=fields.first_collected_field,
    )

    merge_cond = f"updates.{fields.idents_group_id_field} = target.{fields.idents_group_id_field}"
    update_cond_lt = f"updates.{fields.first_collected_field} < target.{fields.first_collected_field}"
    update_cond_gt = f"updates.{fields.first_collected_field} > target.{fields.first_collected_field}"

    (
        delta_table.alias("target")
        .merge(updates_df.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .whenMatchedUpdate(
            update_cond_lt,
            {
                fields.first_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
                fields.last_collected_field: fn.col(
                    f"target.{fields.first_collected_field}",
                ),
            },
        )
        .whenMatchedUpdate(  # update title, description, subscribers, ...
            update_cond_gt,
            {
                fields.last_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
                "title": fn.col("updates.title"),
                "advertiser_category": fn.col("updates.advertiser_category"),
                "description": fn.col("updates.description"),
                "public_description": fn.col("updates.public_description"),
                "display_name": fn.col("updates.display_name"),
                "submit_text": fn.col("updates.submit_text"),
                "submit_link_label": fn.col("updates.submit_link_label"),
                "subscribers": fn.col("updates.subscribers"),
                "accounts_active": fn.col("updates.accounts_active"),
                "active_user_count": fn.col("updates.active_user_count"),
                "url": fn.col("updates.url"),
                "user_flair_text": fn.col("updates.user_flair_text"),
            },
        )
        .execute()
    )


def guilds_upsert_to_delta(
    delta_table: DeltaTable, strategy: Merge_Strategy, input_df: DataFrame
) -> None:
    updates_df = merge_drop_duplicates(
        input_df=input_df.withColumnRenamed(
            fields.bronze_timestamp_field, fields.first_collected_field
        ).withColumn(fields.last_collected_field, fn.expr("cast(null as timestamp)")),
        strategy=strategy,
        id_field=fields.idents_group_id_field,
        strategy_field=fields.first_collected_field,
    )

    merge_cond = f"updates.{fields.idents_group_id_field} = target.{fields.idents_group_id_field}"
    update_cond_lt = f"updates.{fields.first_collected_field} < target.{fields.first_collected_field}"
    update_cond_gt = f"updates.{fields.first_collected_field} > target.{fields.first_collected_field}"

    (
        delta_table.alias("target")
        .merge(updates_df.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .whenMatchedUpdate(
            update_cond_lt,
            {
                fields.first_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
                fields.last_collected_field: fn.col(
                    f"target.{fields.first_collected_field}",
                ),
            },
        )
        .whenMatchedUpdate(
            update_cond_gt,
            {
                fields.last_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
                "name": fn.col("updates.name"),
                "owner": fn.col("updates.owner"),
                "owner_id": fn.col("updates.owner_id"),
                "permissions": fn.col("updates.permissions"),
                "features": fn.col("updates.features"),
                "approximate_member_count": fn.col("updates.approximate_member_count"),
                "approximate_presence_count": fn.col(
                    "updates.approximate_presence_count"
                ),
                "description": fn.col("updates.description"),
            },
        )
        .execute()
    )


# arguments
#   delta_table: delta table
#   strategy: merge strategy
# return: merge function (suitable for `foreachBatch`)
def submissions_upsert_to_delta(
    delta_table: DeltaTable, strategy: Merge_Strategy, input_df: DataFrame
) -> None:
    updates_df = merge_drop_duplicates(
        input_df=input_df.withColumnRenamed(
            fields.bronze_timestamp_field, fields.first_collected_field
        ).withColumn(fields.last_collected_field, fn.expr("cast(null as timestamp)")),
        strategy=strategy,
        id_field=fields.idents_doc_id_field,
        strategy_field=fields.first_collected_field,
    )
    merge_cond = (
        f"updates.{fields.idents_group_id_field} = target.{fields.idents_group_id_field}"
        f" and updates.{fields.idents_doc_id_field} = target.{fields.idents_doc_id_field}"
    )
    update_cond_lt = f"updates.{fields.first_collected_field} < target.{fields.first_collected_field}"
    update_cond_gt = f"updates.{fields.first_collected_field} > target.{fields.first_collected_field}"

    (
        delta_table.alias("target")
        .merge(updates_df.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .whenMatchedUpdate(  # keep original title/text
            update_cond_lt,
            {
                fields.first_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
                fields.last_collected_field: fn.col(
                    f"target.{fields.first_collected_field}",
                ),
                fields.reddit_submissions_title: fn.col(
                    f"updates.{fields.reddit_submissions_title}",
                ),
                fields.reddit_submissions_text: fn.col(
                    f"updates.{fields.reddit_submissions_text}",
                ),
            },
        )
        .whenMatchedUpdate(
            update_cond_gt,
            {
                fields.last_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
                "locked": fn.col("updates.locked"),
                "awarders": fn.col("updates.awarders"),
                "all_awardings": fn.col("updates.all_awardings"),
                "archived": fn.col("updates.archived"),
                "likes": fn.col("updates.likes"),
                "removed_by": fn.col("updates.removed_by"),
                "banned_by": fn.col("updates.banned_by"),
                "banned_at_utc": fn.col("updates.banned_at_utc"),
                "gildings": fn.col("updates.gildings"),
                "upvote_ratio": fn.col("updates.upvote_ratio"),
                "ups": fn.col("updates.ups"),
                "downs": fn.col("updates.downs"),
                "score": fn.col("updates.score"),
                "edited": fn.col("updates.edited"),
                "view_count": fn.col("updates.view_count"),
                "num_comments": fn.col("updates.num_comments"),
                "subreddit_subscribers": fn.col("updates.subreddit_subscribers"),
                "num_crossposts": fn.col("updates.num_crossposts"),
                "source_keys": fn.col("updates.source_keys"),
            },
        )
        .execute()
    )


# arguments
#   delta_table: delta table
#   strategy: merge strategy
# return: merge function (suitable for `foreachBatch`)
def channels_upsert_to_delta(
    delta_table: DeltaTable, strategy: Merge_Strategy, input_df: DataFrame
) -> None:
    updates_df = merge_drop_duplicates(
        input_df=input_df.withColumnRenamed(
            fields.bronze_timestamp_field, fields.first_collected_field
        ).withColumn(fields.last_collected_field, fn.expr("cast(null as timestamp)")),
        strategy=strategy,
        id_field=fields.idents_doc_id_field,
        strategy_field=fields.first_collected_field,
    )
    merge_cond = (
        f"updates.{fields.idents_doc_id_field} = target.{fields.idents_doc_id_field}"
    )
    update_cond_lt = f"updates.{fields.first_collected_field} < target.{fields.first_collected_field}"
    update_cond_gt = f"updates.{fields.first_collected_field} > target.{fields.first_collected_field}"

    (
        delta_table.alias("target")
        .merge(updates_df.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .whenMatchedUpdate(  # keep original title/text
            update_cond_lt,
            {
                fields.first_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
                fields.last_collected_field: fn.col(
                    f"target.{fields.first_collected_field}",
                ),
            },
        )
        .whenMatchedUpdate(
            update_cond_gt,
            {
                fields.last_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
                "type": fn.col("updates.type"),
                "position": fn.col("updates.position"),
                "permission_overwrites": fn.col("updates.permission_overwrites"),
                "name": fn.col("updates.name"),
                "topic": fn.col("updates.topic"),
                "nsfw": fn.col("updates.nsfw"),
                "owner_id": fn.col("updates.owner_id"),
                "rate_limit_per_user": fn.col("updates.rate_limit_per_user"),
                "icon": fn.col("updates.icon"),
                "recipients": fn.col("updates.recipients"),
                "message_count": fn.col("updates.message_count"),
                "member_count": fn.col("updates.member_count"),
                "thread_metadata": fn.col("updates.thread_metadata"),
                "member": fn.col("updates.member"),
                "permissions": fn.col("updates.permissions"),
                "flags": fn.col("updates.flags"),
                "total_message_sent": fn.col("updates.total_message_sent"),
                "available_tags": fn.col("updates.available_tags"),
                "applied_tags": fn.col("updates.applied_tags"),
                "default_reaction_emoji": fn.col("updates.default_reaction_emoji"),
                "default_forum_layout": fn.col("updates.default_forum_layout"),
            },
        )
        .execute()
    )


# arguments
#   delta_table: delta table
#   strategy: merge strategy
# return: merge function (suitable for `foreachBatch`)
def replies_upsert_to_delta(
    delta_table: DeltaTable, strategy: Merge_Strategy, input_df: DataFrame
) -> None:
    updates_df = merge_drop_duplicates(
        input_df=input_df.withColumnRenamed(
            fields.bronze_timestamp_field, fields.first_collected_field
        ).withColumn(fields.last_collected_field, fn.expr("cast(null as timestamp)")),
        strategy=strategy,
        id_field=fields.idents_doc_id_field,
        strategy_field=fields.first_collected_field,
    )
    merge_cond = (
        f"updates.{fields.idents_group_id_field} = target.{fields.idents_group_id_field}"
        f" and updates.{fields.idents_doc_id_field} = target.{fields.idents_doc_id_field}"
    )
    update_cond_lt = f"updates.{fields.first_collected_field} < target.{fields.first_collected_field}"
    update_cond_gt = f"updates.{fields.first_collected_field} > target.{fields.first_collected_field}"

    (
        delta_table.alias("target")
        .merge(updates_df.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .whenMatchedUpdate(  # keep original text
            update_cond_lt,
            {
                fields.first_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
                fields.last_collected_field: fn.col(
                    f"target.{fields.first_collected_field}",
                ),
                fields.reddit_replies_text: fn.col(
                    f"updates.{fields.reddit_replies_text}",
                ),
            },
        )
        .whenMatchedUpdate(
            update_cond_gt,
            {
                fields.last_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
                "archived": fn.col("updates.archived"),
                "all_awardings": fn.col("updates.all_awardings"),
                "associated_award": fn.col("updates.associated_award"),
                "awarders": fn.col("updates.awarders"),
                "banned_at_utc": fn.col("updates.banned_at_utc"),
                "banned_by": fn.col("updates.banned_by"),
                "likes": fn.col("updates.likes"),
                "ups": fn.col("updates.ups"),
                "downs": fn.col("updates.downs"),
                "score": fn.col("updates.score"),
                "edited": fn.col("updates.edited"),
                "submission_uuid": fn.col("updates.submission_uuid"),
                "parent_uuid": fn.col("updates.parent_uuid"),
                "source_keys": fn.col("updates.source_keys"),
            },
        )
        .execute()
    )


# arguments
#   delta_table: delta table
# return: merge function (suitable for `foreachBatch`)
def community_group_upsert_to_delta(
    delta_table: DeltaTable, input_df: DataFrame
) -> None:
    updates_df = input_df.dropDuplicates([fields.idents_group_id_field])

    merge_cond = f"updates.{fields.idents_group_id_field} = target.{fields.idents_group_id_field}"

    (
        delta_table.alias("target")
        .merge(
            updates_df.alias("updates"),
            merge_cond,
        )
        .whenNotMatchedInsertAll()
        .execute()
    )


# arguments
#   delta_table: delta table
#   strategy: merge strategy
# return: merge function (suitable for `foreachBatch`)
def docs_upsert_posts(
    delta_table: DeltaTable, strategy: Merge_Strategy, input_df: DataFrame
) -> None:
    texts_col = fn.struct(
        fn.col(fields.facebook_posts_title).alias("title"),
        fn.col(fields.facebook_posts_text).alias("text"),
    )
    updates_df = merge_drop_duplicates(
        input_df=input_df.select(
            fn.col(fields.idents_doc_id_field),
            fn.col(fields.idents_group_id_field),
            fn.col(fields.source_created_at),
            texts_col.alias(fields.docs_texts),
            fn.col(fields.bronze_timestamp_field).alias(fields.first_collected_field),
            fn.expr("cast(null as timestamp)").alias(fields.last_collected_field),
        ),
        strategy=strategy,
        id_field=fields.idents_doc_id_field,
        strategy_field=fields.first_collected_field,
    )
    merge_cond = (
        f"updates.{fields.idents_group_id_field} = target.{fields.idents_group_id_field}"
        f" and updates.{fields.idents_doc_id_field} = target.{fields.idents_doc_id_field}"
    )

    (
        delta_table.alias("target")
        .merge(updates_df.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .execute()
    )


# NOTE: for the migration the merge logic needs to be more complicated
#       after migration revert to simpler logic
# arguments
#   delta_table: delta table
#   strategy: merge strategy
# return: merge function (suitable for `foreachBatch`)
def docs_upsert_comments(
    delta_table: DeltaTable, strategy: Merge_Strategy, input_df: DataFrame
) -> None:
    texts_col = fn.struct(
        fn.expr("cast(null as string)").alias("title"),
        fn.col(fields.facebook_comments_text).alias("text"),
    )
    updates_df = merge_drop_duplicates(
        input_df=input_df.select(
            fn.col(fields.idents_doc_id_field),
            fn.col(fields.idents_group_id_field),
            fn.col(fields.source_created_at),
            texts_col.alias(fields.docs_texts),
            fn.col(fields.bronze_timestamp_field).alias(fields.first_collected_field),
            fn.expr("cast(null as timestamp)").alias(fields.last_collected_field),
        ),
        strategy=strategy,
        id_field=fields.idents_doc_id_field,
        strategy_field=fields.first_collected_field,
    )
    merge_cond = (
        f"updates.{fields.idents_group_id_field} = target.{fields.idents_group_id_field}"
        f" and updates.{fields.idents_doc_id_field} = target.{fields.idents_doc_id_field}"
    )

    (
        delta_table.alias("target")
        .merge(updates_df.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .execute()
    )


# arguments
#   delta_table: delta table
#   strategy: merge strategy
# return: merge function (suitable for `foreachBatch`)
def docs_upsert_submissions(
    delta_table: DeltaTable, strategy: Merge_Strategy, input_df: DataFrame
) -> None:
    texts_col = fn.struct(
        fn.col(fields.reddit_submissions_title).alias("title"),
        fn.col(fields.reddit_submissions_text).alias("text"),
    )
    updates_df = merge_drop_duplicates(
        input_df=input_df.select(
            fn.col(fields.idents_doc_id_field),
            fn.col(fields.idents_group_id_field),
            fn.col(fields.source_created_at),
            texts_col.alias(fields.docs_texts),
            fn.col(fields.bronze_timestamp_field).alias(fields.first_collected_field),
            fn.expr("cast(null as timestamp)").alias(fields.last_collected_field),
        ),
        strategy=strategy,
        id_field=fields.idents_doc_id_field,
        strategy_field=fields.first_collected_field,
    )
    merge_cond = (
        f"updates.{fields.idents_group_id_field} = target.{fields.idents_group_id_field}"
        f" and updates.{fields.idents_doc_id_field} = target.{fields.idents_doc_id_field}"
    )

    (
        delta_table.alias("target")
        .merge(updates_df.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .execute()
    )


# arguments
#   delta_table: delta table
#   strategy: merge strategy
# return: merge function (suitable for `foreachBatch`)
def docs_upsert_replies(
    delta_table: DeltaTable, strategy: Merge_Strategy, input_df: DataFrame
) -> None:
    texts_col = fn.struct(
        fn.expr("cast(null as string)").alias("title"),
        fn.col(fields.reddit_replies_text).alias("text"),
    )
    updates_df = merge_drop_duplicates(
        input_df=input_df.select(
            fn.col(fields.idents_doc_id_field),
            fn.col(fields.idents_group_id_field),
            fn.col(fields.source_created_at),
            texts_col.alias(fields.docs_texts),
            fn.col(fields.bronze_timestamp_field).alias(fields.first_collected_field),
            fn.expr("cast(null as timestamp)").alias(fields.last_collected_field),
        ),
        strategy=strategy,
        id_field=fields.idents_doc_id_field,
        strategy_field=fields.first_collected_field,
    )
    merge_cond = (
        f"updates.{fields.idents_group_id_field} = target.{fields.idents_group_id_field}"
        f" and updates.{fields.idents_doc_id_field} = target.{fields.idents_doc_id_field}"
    )

    (
        delta_table.alias("target")
        .merge(updates_df.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .execute()
    )


# arguments
#   delta_table: delta table
#   strategy: merge strategy
# return: merge function (suitable for `foreachBatch`)
def docs_upsert_discord(
    delta_table: DeltaTable, strategy: Merge_Strategy, input_df: DataFrame
) -> None:
    texts_col = fn.struct(
        fn.expr("cast(null as string)").alias("title"),
        fn.col("content").alias("text"),
    )
    updates_df = merge_drop_duplicates(
        input_df=input_df.select(
            fn.col(fields.idents_doc_id_field),
            fn.col(fields.idents_group_id_field),
            fn.col(fields.source_created_at),
            texts_col.alias(fields.docs_texts),
            fn.col(fields.bronze_timestamp_field).alias(fields.first_collected_field),
            fn.expr("cast(null as timestamp)").alias(fields.last_collected_field),
        ),
        strategy=strategy,
        id_field=fields.idents_doc_id_field,
        strategy_field=fields.first_collected_field,
    )
    merge_cond = (
        f"updates.{fields.idents_group_id_field} = target.{fields.idents_group_id_field}"
        f" and updates.{fields.idents_doc_id_field} = target.{fields.idents_doc_id_field}"
    )

    (
        delta_table.alias("target")
        .merge(updates_df.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .execute()
    )


# arguments
#   delta_table: delta table
#   strategy: merge strategy
# return: merge function (suitable for `foreachBatch`)
def messages_upsert_to_delta(
    delta_table: DeltaTable, strategy: Merge_Strategy, input_df: DataFrame
) -> None:
    updates_df = merge_drop_duplicates(
        input_df=input_df.withColumnRenamed(
            fields.bronze_timestamp_field, fields.first_collected_field
        ).withColumn(fields.last_collected_field, fn.expr("cast(null as timestamp)")),
        strategy=strategy,
        id_field=fields.idents_doc_id_field,
        strategy_field=fields.first_collected_field,
    )
    merge_cond = (
        f"updates.{fields.idents_doc_id_field} = target.{fields.idents_doc_id_field}"
    )
    update_cond_lt = f"updates.{fields.first_collected_field} < target.{fields.first_collected_field}"
    update_cond_gt = f"updates.{fields.first_collected_field} > target.{fields.first_collected_field}"

    (
        delta_table.alias("target")
        .merge(updates_df.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .whenMatchedUpdate(  # keep original content
            update_cond_lt,
            {
                fields.first_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
                fields.last_collected_field: fn.col(
                    f"target.{fields.first_collected_field}",
                ),
                "content": fn.col("updates.content"),
            },
        )
        .whenMatchedUpdate(
            update_cond_gt,
            {
                fields.last_collected_field: fn.col(
                    f"updates.{fields.first_collected_field}",
                ),
                "edited_at": fn.col("updates.edited_at"),
                "author": fn.col("updates.author"),
                "tts": fn.col("updates.tts"),
                "mention_everyone": fn.col("updates.mention_everyone"),
                "mentions": fn.col("updates.mentions"),
                "mention_channels": fn.col("updates.mention_channels"),
                "embeds": fn.col("updates.embeds"),
                "reactions": fn.col("updates.reactions"),
                "pinned": fn.col("updates.pinned"),
                "type": fn.col("updates.type"),
                "flags": fn.col("updates.flags"),
                "thread": fn.col("updates.thread"),
                "sticker_items": fn.col("updates.sticker_items"),
                "position": fn.col("updates.position"),
                "poll": fn.col("updates.poll"),
                "call": fn.col("updates.call"),
            },
        )
        .execute()
    )


# arguments
#   delta_table: delta table
# return: merge function (suitable for `foreachBatch`)
def reactions_upsert_to_delta(delta_table: DeltaTable, input_df: DataFrame) -> None:
    updates_df = (
        merge_drop_duplicates_key(
            input_df,
            fn.format_string(
                "%s%s%s", fn.col("doc_uuid"), fn.col("user.id"), fn.col("emoji.name")
            ),
        )
        .withColumnRenamed(fields.bronze_timestamp_field, fields.first_collected_field)
        .withColumn(fields.last_collected_field, fn.expr("cast(null as timestamp)"))
    )
    merge_cond = (
        f"updates.{fields.idents_doc_id_field} = target.{fields.idents_doc_id_field}"
        " and updates.user.id = target.user.id"
        " and updates.emoji.name = target.emoji.name"
    )

    (
        delta_table.alias("target")
        .merge(updates_df.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .execute()
    )


# return: merge function (suitable for `foreachBatch`)
def members_upsert_to_delta(delta_table: DeltaTable, input_df: DataFrame) -> None:
    updates_df = (
        merge_drop_duplicates_key(
            input_df,
            fn.format_string("%s%s", fn.col("group_id"), fn.col("user.id")),
        )
        .withColumnRenamed(fields.bronze_timestamp_field, fields.first_collected_field)
        .withColumn(fields.last_collected_field, fn.expr("cast(null as timestamp)"))
    )
    merge_cond = (
        f"updates.{fields.idents_group_id_field} = target.{fields.idents_group_id_field}"
        " and updates.user.id = target.user.id"
    )

    (
        delta_table.alias("target")
        .merge(updates_df.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .execute()
    )


# arguments
#   delta_table: delta table
#   updates: dataframe with update delta (in schema of delta_table)
def daily_ingestion_counts_upsert_to_delta(
    delta_table: DeltaTable, updates: DataFrame, batch_id: int
) -> None:
    merge_cond = (
        f"updates.{fields.idents_group_id_field} = target.{fields.idents_group_id_field}"
        f" and updates.{fields.collected_day_field} = target.{fields.collected_day_field}"
        f" and updates.{fields.statement_type_field} = target.{fields.statement_type_field}"
        f" and updates.{fields.batch_id_field} = target.{fields.batch_id_field}"
    )
    sum_col = fn.col(f"updates.{fields.count_field}") + fn.col(
        f"target.{fields.count_field}"
    )
    (
        delta_table.alias("target")
        .merge(
            updates.withColumn(fields.batch_id_field, fn.lit(batch_id)).alias(
                "updates"
            ),
            merge_cond,
        )
        .whenNotMatchedInsertAll()
        .whenMatchedUpdate(set={fields.count_field: sum_col})
        .execute()
    )


def job_run_ingestion_counts_upsert_to_delta(
    delta_table: DeltaTable, updates: DataFrame, batch_id: int
) -> None:
    merge_cond = (
        f"updates.{fields.idents_group_id_field} = target.{fields.idents_group_id_field}"
        f" and updates.{fields.job_run_id_field} = target.{fields.job_run_id_field}"
        f" and updates.{fields.statement_type_field} = target.{fields.statement_type_field}"
        f" and updates.{fields.batch_id_field} = target.{fields.batch_id_field}"
    )
    (
        delta_table.alias("target")
        .merge(
            updates.withColumn(fields.batch_id_field, fn.lit(batch_id)).alias(
                "updates"
            ),
            merge_cond,
        )
        .whenNotMatchedInsertAll()
        .execute()
    )


# similar to favor_earliest:
#   on lt cond use favor_update
#   on gt cond use favor_target
# similar to favor_earliest_text:
#   on lt cond use favor_update
#   ignore gt cond
# similar to use_most_recent:
#   ignore lt cond
#   on gt cond set to update value
# similar to merge_author:
#   on lt cond use merge_author_favor_update
#   on gt cond use merge_author_favor_target
# arguments:
#   delta_table - submissions gold table
#   updates - assumes table has been filtered and enriched -- has the same
#             schema as the submissions gold table.
def merge_canonical_reddit(
    delta_table: DeltaTable, updates: DataFrame, schema: Literal["submission", "reply"]
) -> None:
    merge_cond = "updates.source_keys.name = target.source_keys.name"
    lt_cond = (
        "updates.ingested_on < target.ingested_on"
        " or ("
        "updates.ingested_on = target.ingested_on"
        " AND updates.ingest_version < target.ingest_version"
        ")"
    )
    gt_cond = (
        "updates.ingested_on > target.ingested_on"
        " or ("
        "updates.ingested_on = target.ingested_on"
        " AND updates.ingest_version > target.ingest_version"
        ")"
    )

    def favor_update(field: str) -> Column:
        return fn.coalesce(
            fn.col(f"updates.{field}"),
            fn.col(f"target.{field}"),
        )

    def favor_target(field: str) -> Column:
        return fn.coalesce(
            fn.col(f"target.{field}"),
            fn.col(f"updates.{field}"),
        )

    def merge_author_favor_update() -> Column:
        return fn.coalesce(
            null_if_deleted(fn.col("updates.source_keys.author")),
            null_if_deleted(fn.col("target.source_keys.author")),
            fn.col("updates.source_keys.author"),
            fn.col("target.source_keys.author"),
        )

    def merge_author_favor_target() -> Column:
        return fn.coalesce(
            null_if_deleted(fn.col("target.source_keys.author")),
            null_if_deleted(fn.col("updates.source_keys.author")),
            fn.col("target.source_keys.author"),
            fn.col("updates.source_keys.author"),
        )

    source_keys, most_recent_cols, text_cols, rem = partition_with_remaining(
        updates.columns,
        ["source_keys"],
        (
            [
                "num_comments",
                "num_crossposts",
                "score",
                "ups",
                "downs",
                "upvote_ratio",
                "scores_collected_on",
            ]
            if schema == "submission"
            else [
                "num_comments",
                "score",
                "ups",
                "downs",
                "scores_collected_on",
            ]
        ),
        ["text_version", "text"],
    )

    set_lt_cond: dict[str, Union[str, Column]] = {
        **{
            col: (
                fn.struct(
                    favor_update("source_keys.name").alias("name"),
                    merge_author_favor_update().alias("author"),
                    favor_update("source_keys.author_fullname").alias(
                        "author_fullname"
                    ),
                    favor_update("source_keys.subreddit_id").alias("subreddit_id"),
                )
                if schema == "submission"
                else fn.struct(
                    favor_update("source_keys.name").alias("name"),
                    favor_update("source_keys.submission_id").alias("submission_id"),
                    favor_update("source_keys.parent_id").alias("parent_id"),
                    merge_author_favor_update().alias("author"),
                    favor_update("source_keys.author_fullname").alias(
                        "author_fullname"
                    ),
                    favor_update("source_keys.subreddit_id").alias("subreddit_id"),
                )
            )
            for col in source_keys
        },
        **{col: favor_update(col) for col in text_cols},
        **{col: favor_update(col) for col in rem},
    }

    set_gt_cond: dict[str, Union[str, Column]] = {
        **{
            col: (
                fn.struct(
                    favor_target("source_keys.name").alias("name"),
                    merge_author_favor_target().alias("author"),
                    favor_target("source_keys.author_fullname").alias(
                        "author_fullname"
                    ),
                    favor_target("source_keys.subreddit_id").alias("subreddit_id"),
                )
                if schema == "submission"
                else fn.struct(
                    favor_target("source_keys.name").alias("name"),
                    favor_target("source_keys.submission_id").alias("submission_id"),
                    favor_target("source_keys.parent_id").alias("parent_id"),
                    merge_author_favor_target().alias("author"),
                    favor_target("source_keys.author_fullname").alias(
                        "author_fullname"
                    ),
                    favor_target("source_keys.subreddit_id").alias("subreddit_id"),
                )
            )
            for col in source_keys
        },
        **{col: favor_update(col) for col in most_recent_cols},
        **{col: favor_target(col) for col in rem},
    }

    (
        delta_table.alias("target")
        .merge(updates.alias("updates"), merge_cond)
        .whenNotMatchedInsertAll()
        .whenMatchedUpdate(
            condition=lt_cond,
            set=set_lt_cond,
        )
        .whenMatchedUpdate(
            condition=gt_cond,
            set=set_gt_cond,
        )
        .execute()
    )
