from typing import Callable
import functools

from pyspark.sql import Column
import pyspark.sql.functions as fn
from pyspark.sql.types import BinaryType, IntegerType

import trend_etl.schema.types as types


def binary_coalesce(col: Column) -> Column:
    return fn.coalesce(col, fn.lit("").cast(BinaryType()))


def direct_to_binary(col: Column) -> Column:
    return binary_coalesce(col.cast(BinaryType()))


def boolean_to_binary(col: Column) -> Column:
    return direct_to_binary(col.cast(IntegerType()))


def array_to_binary(to_binary: Callable[[Column], Column], col: Column) -> Column:
    return binary_coalesce(
        fn.aggregate(
            col,
            fn.lit("").cast(BinaryType()),
            lambda acc, elem: fn.concat(acc, binary_coalesce(to_binary(elem))),
        )
    )


array_direct_to_binary = functools.partial(array_to_binary, direct_to_binary)


def facebook_reaction_to_binary(col: Column) -> Column:
    return direct_to_binary(col.getField("summary").getField("total_count"))


def facebook_parent_to_binary(col: Column) -> Column:
    return direct_to_binary(col.getField("id"))


def reddit_comment_contribution_settings_to_binary(col: Column) -> Column:
    return array_direct_to_binary(col.getField("allowed_media_types"))


def reddit_all_awardings_to_binary(col: Column) -> Column:
    return array_to_binary(
        lambda elem: fn.concat(
            direct_to_binary(elem.getField("award_sub_type")),
            direct_to_binary(elem.getField("award_type")),
            direct_to_binary(elem.getField("awardings_required_to_grant_benefits")),
            direct_to_binary(elem.getField("coin_price")),
            direct_to_binary(elem.getField("coin_reward")),
            direct_to_binary(elem.getField("count")),
            direct_to_binary(elem.getField("days_of_drip_extension")),
            direct_to_binary(elem.getField("days_of_premium")),
            direct_to_binary(elem.getField("description")),
            direct_to_binary(elem.getField("end_date")),
            direct_to_binary(elem.getField("giver_coin_reward")),
            direct_to_binary(elem.getField("icon_format")),
            direct_to_binary(elem.getField("icon_height")),
            direct_to_binary(elem.getField("icon_url")),
            direct_to_binary(elem.getField("icon_width")),
            direct_to_binary(elem.getField("id")),
            boolean_to_binary(elem.getField("is_enabled")),
            boolean_to_binary(elem.getField("is_new")),
            direct_to_binary(elem.getField("name")),
            direct_to_binary(elem.getField("penny_donate")),
            direct_to_binary(elem.getField("penny_price")),
            direct_to_binary(elem.getField("start_date")),
            direct_to_binary(elem.getField("static_icon_height")),
            direct_to_binary(elem.getField("static_icon_url")),
            direct_to_binary(elem.getField("static_icon_width")),
            direct_to_binary(elem.getField("sticky_duration_seconds")),
            direct_to_binary(elem.getField("subreddit_coin_reward")),
            direct_to_binary(elem.getField("subreddit_id")),
        ),
        col,
    )


def reddit_link_flair_richtext_to_binary(col: Column) -> Column:
    return array_to_binary(
        lambda elem: fn.concat(
            direct_to_binary(elem.getField("e")),
            direct_to_binary(elem.getField("t")),
        ),
        col,
    )


def reddit_author_flair_richtext_to_binary(col: Column) -> Column:
    return array_to_binary(
        lambda elem: fn.concat(
            direct_to_binary(elem.getField("a")),
            direct_to_binary(elem.getField("e")),
            direct_to_binary(elem.getField("t")),
            direct_to_binary(elem.getField("u")),
        ),
        col,
    )


def reddit_gildings_to_binary(col: Column) -> Column:
    return fn.concat(
        direct_to_binary(col.getField("gid_1")),
        direct_to_binary(col.getField("gid_2")),
        direct_to_binary(col.getField("gid_3")),
    )


def discord_available_tags_to_binary(col: Column) -> Column:
    return array_to_binary(
        lambda elem: fn.concat(
            direct_to_binary(elem.getField("id")),
            direct_to_binary(elem.getField("name")),
            boolean_to_binary(elem.getField("moderated")),
            direct_to_binary(elem.getField("emoji_id")),
            direct_to_binary(elem.getField("emoji_name")),
        ),
        col,
    )


def discord_permission_overwrites_to_binary(col: Column) -> Column:
    return array_to_binary(
        lambda elem: fn.concat(
            direct_to_binary(elem.getField("id")),
            direct_to_binary(elem.getField("type")),
            direct_to_binary(elem.getField("allow")),
            direct_to_binary(elem.getField("deny")),
        ),
        col,
    )


def discord_thread_metadata_to_binary(col: Column) -> Column:
    return fn.concat(
        boolean_to_binary(col.getField("archived")),
        direct_to_binary(col.getField("auto_archive_duration")),
        direct_to_binary(col.getField("archive_timestamp")),
        boolean_to_binary(col.getField("locked")),
        boolean_to_binary(col.getField("invitable")),
        direct_to_binary(col.getField("create_timestamp")),
    )


def discord_default_reaction_emoji_to_binary(col: Column) -> Column:
    return fn.concat(
        direct_to_binary(col.getField("emoji_id")),
        direct_to_binary(col.getField("emoji_name")),
    )


def discord_embeds_to_binary(col: Column) -> Column:
    return array_to_binary(
        lambda elem: fn.concat(
            direct_to_binary(elem.getField("title")),
            direct_to_binary(elem.getField("type")),
            direct_to_binary(elem.getField("description")),
            direct_to_binary(elem.getField("url")),
            direct_to_binary(elem.getField("timestamp")),
            direct_to_binary(elem.getField("image").getField("url")),
            direct_to_binary(elem.getField("video").getField("url")),
            fn.concat(
                direct_to_binary(elem.getField("author").getField("name")),
                direct_to_binary(elem.getField("author").getField("url")),
            ),
            fn.concat(
                direct_to_binary(elem.getField("provider").getField("name")),
                direct_to_binary(elem.getField("provider").getField("url")),
            ),
        ),
        col,
    )


def discord_message_reference_to_binary(col: Column) -> Column:
    return fn.concat(
        direct_to_binary(col.getField("type")),
        direct_to_binary(col.getField("channel_id")),
        direct_to_binary(col.getField("message_id")),
        direct_to_binary(col.getField("guild_id")),
    )


def discord_thread_to_binary(col: Column) -> Column:
    return fn.concat(
        direct_to_binary(col.getField("id")),
        direct_to_binary(col.getField("type")),
        direct_to_binary(col.getField("name")),
        direct_to_binary(col.getField("guild_id")),
    )


def discord_sticker_items_to_binary(col: Column) -> Column:
    return array_to_binary(
        lambda elem: fn.concat(
            direct_to_binary(elem.getField("id")),
            direct_to_binary(elem.getField("name")),
            direct_to_binary(elem.getField("format_type")),
        ),
        col,
    )


def discord_poll_to_binary(col: Column) -> Column:
    return fn.concat(
        fn.concat(
            direct_to_binary(
                col.getField("question").getField("text"),
            ),
            fn.concat(
                direct_to_binary(
                    col.getField("question").getField("emoji").getField("id")
                ),
                direct_to_binary(
                    col.getField("question").getField("emoji").getField("name")
                ),
            ),
        ),
        array_to_binary(
            lambda elem: fn.concat(
                direct_to_binary(elem.getField("answer_id")),
                fn.concat(
                    direct_to_binary(elem.getField("poll_media").getField("text")),
                    direct_to_binary(
                        elem.getField("poll_media").getField("emoji").getField("id")
                    ),
                    direct_to_binary(
                        elem.getField("poll_media").getField("emoji").getField("name")
                    ),
                ),
            ),
            col.getField("answers"),
        ),
        direct_to_binary(
            col.getField("expiry"),
        ),
        boolean_to_binary(
            col.getField("allow_multiselect"),
        ),
        fn.concat(
            boolean_to_binary(
                col.getField("results").getField("is_finalized"),
            ),
            array_to_binary(
                lambda elem: fn.concat(
                    direct_to_binary(elem.getField("id")),
                    direct_to_binary(elem.getField("count")),
                    boolean_to_binary(elem.getField("me_voted")),
                ),
                col.getField("results").getField("answer_counts"),
            ),
        ),
    )


def discord_call_to_binary(col: Column) -> Column:
    return fn.concat(
        array_to_binary(
            direct_to_binary,
            col.getField("participants"),
        ),
        direct_to_binary(
            col.getField("ended_timestamp"),
        ),
    )


def discord_emoji_to_binary(col: Column) -> Column:
    return fn.concat(
        direct_to_binary(col.getField("id")),
        direct_to_binary(col.getField("name")),
    )


def discord_avatar_decoration_data_to_binary(col: Column) -> Column:
    return fn.concat(
        direct_to_binary(col.getField("asset")),
        direct_to_binary(col.getField("sku_id")),
    )


def ingest_version(struct_data: Column, canonical: types.Canonical) -> Column:
    return fn.md5(
        fn.concat(
            *[
                to_binary(struct_data.getField(field_key))
                for field_key, to_binary in canonical
            ]
        )
    )
