import unittest
import os
import shutil
from datetime import datetime

from pyspark.sql import SparkSession
import pyspark.sql.functions as fn
from delta.tables import DeltaTable

from trend_etl.test_utils import getspark, assert_dataframes_equal
from trend_etl.functions.delta import merge_earliest_version, earliest_version


class TestDeltaEarliestMerge(unittest.TestCase):
    spark: SparkSession
    schema: str
    initial_data: list
    next_data: list
    after_initial_data: list
    after_next_data: list

    @classmethod
    def setUpClass(cls) -> None:
        cls.spark = getspark()
        cls.schema = (
            "fruit struct<fruit string,picked_on int,number_collected int>"
            ",ingested_on timestamp"
            ",ingest_version string"
        )
        cls.initial_data = [
            (("apple", 1, 42), datetime(2024, 12, 24, 0, 0, 0), "v1"),
            (("apple", 2, 42), datetime(2024, 12, 24, 0, 0, 0), "v2"),
            (("lemon", 1, 666), datetime(2024, 12, 24, 10, 30, 0), "v1"),
            (("lemon", 42, 1337), datetime(2025, 12, 24, 10, 30, 0), "v2"),
            (("pineapple", 1, 1), datetime(2025, 12, 24, 11, 0, 0), "v1"),
        ]
        cls.next_data = [
            (("apple", 3, 69), datetime(2024, 12, 24, 0, 0, 0), "v3"),
            (("lemon", 420, 2001), datetime(2025, 12, 25, 10, 30, 0), "v0"),
            (("pineapple", 12, 111), datetime(2024, 12, 24, 11, 0, 0), "v1"),
        ]
        cls.after_initial_data = [
            (("apple", 1, 42), datetime(2024, 12, 24, 0, 0, 0), "v1"),
            (("lemon", 1, 666), datetime(2024, 12, 24, 10, 30, 0), "v1"),
            (("pineapple", 1, 1), datetime(2025, 12, 24, 11, 0, 0), "v1"),
        ]
        cls.after_next_data = [
            (("apple", 1, 42), datetime(2024, 12, 24, 0, 0, 0), "v1"),
            (("lemon", 1, 666), datetime(2024, 12, 24, 10, 30, 0), "v1"),
            (("pineapple", 12, 111), datetime(2024, 12, 24, 11, 0, 0), "v1"),
        ]

    def setUp(self):
        # Set up a temporary directory for the Delta table
        self.temp_dir = "/tmp/delta-table"

        # Ensure path is clean
        if os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)

        bootstrap_table = self.spark.createDataFrame([], schema=self.schema)
        bootstrap_table.write.format("delta").save(self.temp_dir)
        self.delta_table = DeltaTable.forPath(self.spark, self.temp_dir)

    def tearDown(self):
        # Clean up the temporary directory after each test
        if os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)

    def test_merge_earliest_version_initial(self) -> None:
        spark = self.spark

        initial_df = spark.createDataFrame(self.initial_data, schema=self.schema)
        updates = earliest_version(initial_df, fn.col("fruit.fruit"))
        merge_earliest_version(
            self.delta_table,
            updates,
            ["fruit.fruit"],
        )

        got = spark.read.format("delta").load(self.temp_dir)
        want = spark.createDataFrame(self.after_initial_data, schema=self.schema)

        assert_dataframes_equal(got, want)

        next_df = spark.createDataFrame(self.next_data, schema=self.schema)
        updates = earliest_version(next_df, fn.col("fruit.fruit"))
        merge_earliest_version(
            self.delta_table,
            updates,
            ["fruit.fruit"],
        )

        got = spark.read.format("delta").load(self.temp_dir)
        want = spark.createDataFrame(self.after_next_data, schema=self.schema)

        assert_dataframes_equal(got, want)
