import unittest
import os
import shutil

from pyspark.sql import SparkSession, Row
import pyspark.sql.functions as fn
from pyspark.sql.types import StructType
from delta.tables import DeltaTable

from trend_etl.test_utils import getspark, schema_df, get_df
from trend_etl.functions.delta import merge_canonical_reddit


class TestDeltaMergeSubmissions(unittest.TestCase):
    spark: SparkSession
    schema: StructType

    @classmethod
    def setUpClass(cls) -> None:
        cls.spark = getspark()

    def setUp(self):
        # Set up a temporary directory for the Delta table
        self.temp_dir = "/tmp/delta-table"

        # Ensure path is clean
        if os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)

        self.schema = schema_df(
            "test_data", "merge_canonical", "submissions", "gold_schema.json"
        )

        bootstrap_table = self.spark.createDataFrame([], schema=self.schema)
        bootstrap_table.write.format("delta").save(self.temp_dir)
        self.delta_table = DeltaTable.forPath(self.spark, self.temp_dir)

    def tearDown(self):
        # Clean up the temporary directory after each test
        if os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)

    def test_earlier_text_1(self) -> None:
        updates = get_df(
            self.spark,
            self.schema,
            "test_data",
            "merge_canonical",
            "submissions",
            "earlier_text_example.jsonl",
        )

        earlier_record = updates.where(
            fn.col("ingest_version") == "00210297cc36723a21a2047c7daa83d8"
        )
        later_record = updates.where(
            fn.col("ingest_version") == "a0210297cc36723a21a2047c7daa83d8"
        )

        merge_canonical_reddit(self.delta_table, earlier_record, "submission")
        merge_canonical_reddit(self.delta_table, later_record, "submission")

        got = (
            self.spark.read.format("delta")
            .load(self.temp_dir)
            .selectExpr(
                "ingest_version",
                "text",
            )
            .collect()
        )

        self.assertEqual(
            got,
            [
                Row(
                    ingest_version="00210297cc36723a21a2047c7daa83d8",
                    text="hello",
                )
            ],
        )

    def test_earlier_text_2(self) -> None:
        updates = get_df(
            self.spark,
            self.schema,
            "test_data",
            "merge_canonical",
            "submissions",
            "earlier_text_example.jsonl",
        )

        earlier_record = updates.where(
            fn.col("ingest_version") == "00210297cc36723a21a2047c7daa83d8"
        )
        later_record = updates.where(
            fn.col("ingest_version") == "a0210297cc36723a21a2047c7daa83d8"
        )

        merge_canonical_reddit(self.delta_table, later_record, "submission")
        merge_canonical_reddit(self.delta_table, earlier_record, "submission")

        got = (
            self.spark.read.format("delta")
            .load(self.temp_dir)
            .selectExpr("ingest_version", "text")
            .collect()
        )

        self.assertEqual(
            got,
            [
                Row(
                    ingest_version="00210297cc36723a21a2047c7daa83d8",
                    text="hello",
                )
            ],
        )

    def test_latest_score_1(self) -> None:
        updates = get_df(
            self.spark,
            self.schema,
            "test_data",
            "merge_canonical",
            "submissions",
            "latest_score_example.jsonl",
        )

        earlier_record = updates.where(
            fn.col("ingest_version") == "00210297cc36723a21a2047c7daa83d8"
        )
        later_record = updates.where(
            fn.col("ingest_version") == "a0210297cc36723a21a2047c7daa83d8"
        )

        merge_canonical_reddit(self.delta_table, earlier_record, "submission")
        merge_canonical_reddit(self.delta_table, later_record, "submission")

        got = (
            self.spark.read.format("delta")
            .load(self.temp_dir)
            .selectExpr(
                "ingest_version",
                "upvote_ratio",
                "score",
            )
            .collect()
        )

        self.assertEqual(
            got,
            [
                Row(
                    ingest_version="00210297cc36723a21a2047c7daa83d8",
                    upvote_ratio=0.42,
                    score=6,
                )
            ],
        )

    def test_latest_score_2(self) -> None:
        updates = get_df(
            self.spark,
            self.schema,
            "test_data",
            "merge_canonical",
            "submissions",
            "latest_score_example.jsonl",
        )

        earlier_record = updates.where(
            fn.col("ingest_version") == "00210297cc36723a21a2047c7daa83d8"
        )
        later_record = updates.where(
            fn.col("ingest_version") == "a0210297cc36723a21a2047c7daa83d8"
        )

        merge_canonical_reddit(self.delta_table, later_record, "submission")
        merge_canonical_reddit(self.delta_table, earlier_record, "submission")

        got = (
            self.spark.read.format("delta")
            .load(self.temp_dir)
            .selectExpr(
                "ingest_version",
                "upvote_ratio",
                "score",
            )
            .collect()
        )

        self.assertEqual(
            got,
            [
                Row(
                    ingest_version="00210297cc36723a21a2047c7daa83d8",
                    upvote_ratio=0.42,
                    score=6,
                )
            ],
        )

    def test_author_earliest_deleted_1(self) -> None:
        updates = get_df(
            self.spark,
            self.schema,
            "test_data",
            "merge_canonical",
            "submissions",
            "author_earliest_deleted.jsonl",
        )

        earlier_record = updates.where(
            fn.col("ingest_version") == "00210297cc36723a21a2047c7daa83d8"
        )
        later_record = updates.where(
            fn.col("ingest_version") == "a0210297cc36723a21a2047c7daa83d8"
        )

        merge_canonical_reddit(self.delta_table, earlier_record, "submission")
        merge_canonical_reddit(self.delta_table, later_record, "submission")

        got = (
            self.spark.read.format("delta")
            .load(self.temp_dir)
            .selectExpr("source_keys.*")
            .collect()
        )

        self.assertEqual(
            got,
            [
                Row(
                    name="t3_17m1nnx",
                    author="Impossible_Ad_4064",
                    author_fullname="t2_cjn0wkg2",
                    subreddit_id="t5_28l9xj",
                )
            ],
        )

    def test_author_earliest_deleted_2(self) -> None:
        updates = get_df(
            self.spark,
            self.schema,
            "test_data",
            "merge_canonical",
            "submissions",
            "author_earliest_deleted.jsonl",
        )

        earlier_record = updates.where(
            fn.col("ingest_version") == "00210297cc36723a21a2047c7daa83d8"
        )
        later_record = updates.where(
            fn.col("ingest_version") == "a0210297cc36723a21a2047c7daa83d8"
        )

        merge_canonical_reddit(self.delta_table, later_record, "submission")
        merge_canonical_reddit(self.delta_table, earlier_record, "submission")

        got = (
            self.spark.read.format("delta")
            .load(self.temp_dir)
            .selectExpr("source_keys.*")
            .collect()
        )

        self.assertEqual(
            got,
            [
                Row(
                    name="t3_17m1nnx",
                    author="Impossible_Ad_4064",
                    author_fullname="t2_cjn0wkg2",
                    subreddit_id="t5_28l9xj",
                )
            ],
        )

    def test_author_latest_deleted_1(self) -> None:
        updates = get_df(
            self.spark,
            self.schema,
            "test_data",
            "merge_canonical",
            "submissions",
            "author_latest_deleted.jsonl",
        )

        earlier_record = updates.where(
            fn.col("ingest_version") == "00210297cc36723a21a2047c7daa83d8"
        )
        later_record = updates.where(
            fn.col("ingest_version") == "a0210297cc36723a21a2047c7daa83d8"
        )

        merge_canonical_reddit(self.delta_table, earlier_record, "submission")
        merge_canonical_reddit(self.delta_table, later_record, "submission")

        got = (
            self.spark.read.format("delta")
            .load(self.temp_dir)
            .selectExpr("source_keys.*")
            .collect()
        )

        self.assertEqual(
            got,
            [
                Row(
                    name="t3_17m1nnx",
                    author="Impossible_Ad_4064",
                    author_fullname="t2_cjn0wkg2",
                    subreddit_id="t5_28l9xj",
                )
            ],
        )

    def test_author_latest_deleted_2(self) -> None:
        updates = get_df(
            self.spark,
            self.schema,
            "test_data",
            "merge_canonical",
            "submissions",
            "author_latest_deleted.jsonl",
        )

        earlier_record = updates.where(
            fn.col("ingest_version") == "00210297cc36723a21a2047c7daa83d8"
        )
        later_record = updates.where(
            fn.col("ingest_version") == "a0210297cc36723a21a2047c7daa83d8"
        )

        merge_canonical_reddit(self.delta_table, later_record, "submission")
        merge_canonical_reddit(self.delta_table, earlier_record, "submission")

        got = (
            self.spark.read.format("delta")
            .load(self.temp_dir)
            .selectExpr("source_keys.*")
            .collect()
        )

        self.assertEqual(
            got,
            [
                Row(
                    name="t3_17m1nnx",
                    author="Impossible_Ad_4064",
                    author_fullname="t2_cjn0wkg2",
                    subreddit_id="t5_28l9xj",
                )
            ],
        )
