import unittest

from pyspark.sql import SparkSession
from pyspark.sql.functions import (
    col,
    md5,
)

from trend_etl.test_utils import getspark, get_df
from .udfs import annotation_matches_text, double_to_binary


class TestFunctions(unittest.TestCase):
    spark: SparkSession

    @classmethod
    def setUpClass(cls) -> None:
        cls.spark = getspark()

    def test_annotation_matches_text(self) -> None:
        docs = get_df(
            self.__class__.spark,
            "doc_uuid string, texts struct<title string, text string>",
            "test_data",
            "docs.jsonl",
        )
        nlp_results = get_df(
            self.__class__.spark,
            "doc_uuid string, annotation string",
            "test_data",
            "historical_nlp_results.jsonl",
        )

        joined = nlp_results.join(docs, "doc_uuid")

        self.assertEqual(joined.count(), nlp_results.count(), "count mismatch")

        failed_matches = [
            row["doc_uuid"]
            for row in joined.where(
                ~annotation_matches_text(col("texts.text"), col("annotation"))
            )
            .selectExpr("doc_uuid")
            .collect()
        ]

        self.assertEqual(failed_matches, ["doc-fail-1"])

    def test_float_to_binary(self) -> None:
        data = [(3.14,)]
        input_df = self.__class__.spark.createDataFrame(data, schema="float double")

        got_df = input_df.withColumn(
            "ingest_version", md5(double_to_binary(col("float")))
        )

        # $ echo -n -e "\x40\x09\x1e\xb8\x51\xeb\x85\x1f" | md5
        # 32200b8781d6e8f31543da4cf19ff307

        self.assertEqual(got_df.count(), 1, "count mismatch")
        self.assertEqual(
            got_df.collect()[0]["ingest_version"],
            "32200b8781d6e8f31543da4cf19ff307",
            "doc version mismatch",
        )

    def test_null_float_to_binary(self) -> None:
        data = [(None,)]
        input_df = self.__class__.spark.createDataFrame(data, schema="float double")

        got_df = input_df.withColumn(
            "ingest_version", md5(double_to_binary(col("float")))
        )

        # $ echo -n -e "" | md5
        # d41d8cd98f00b204e9800998ecf8427e

        self.assertEqual(got_df.count(), 1, "count mismatch")
        self.assertEqual(
            got_df.collect()[0]["ingest_version"],
            "d41d8cd98f00b204e9800998ecf8427e",
            "doc version mismatch",
        )
