import argparse
from typing import cast

from delta.tables import DeltaTable
from pyspark.sql import SparkSession, DataFrame
import pyspark.sql.functions as fn
from pyspark.sql.streaming import StreamingQuery

from trend_etl.functions.delta import (
    most_recent_version,
    earliest_version,
    merge_most_recent_version,
    merge_earliest_version,
)
import trend_etl.schema.args as arguments

program = "canonicalize_discord"


def write_guilds(
    spark: SparkSession,
    silver_table: str,
    gold_table: str,
    checkpoints_path: str,
) -> StreamingQuery:
    streaming = spark.readStream.format("delta").table(silver_table)

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(spark, gold_table)

        canonical = most_recent_version(batch_df, fn.col("source_keys.id"))

        return merge_most_recent_version(
            delta_table=delta_table, updates=canonical, merge_keys=["source_keys.id"]
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoints_path)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def write_channels(
    spark: SparkSession,
    silver_table: str,
    gold_table: str,
    checkpoints_path: str,
) -> StreamingQuery:
    streaming = spark.readStream.format("delta").table(silver_table)

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(spark, gold_table)

        canonical = most_recent_version(batch_df, fn.col("source_keys.id"))

        return merge_most_recent_version(
            delta_table=delta_table, updates=canonical, merge_keys=["source_keys.id"]
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoints_path)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


# NOTE: need to materialize threads since we need the earliest record version
def write_threads(
    spark: SparkSession,
    silver_table: str,  # uses channels silver table
    gold_table: str,
    checkpoints_path: str,
) -> StreamingQuery:
    streaming = spark.readStream.format("delta").table(silver_table)

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(spark, gold_table)

        canonical = earliest_version(batch_df, fn.col("source_keys.id"))

        return merge_earliest_version(
            delta_table=delta_table, updates=canonical, merge_keys=["source_keys.id"]
        )

    return (
        # filter out non-threads channel records
        streaming.where(fn.col("type").isin(arguments.discord_thread_types))
        .writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoints_path)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def write_members(
    spark: SparkSession,
    silver_table: str,
    gold_table: str,
    checkpoints_path: str,
) -> StreamingQuery:
    streaming = spark.readStream.format("delta").table(silver_table)

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(spark, gold_table)

        canonical = most_recent_version(batch_df, fn.col("source_keys"))

        return merge_most_recent_version(
            delta_table=delta_table,
            updates=canonical,
            merge_keys=["source_keys.guild_id", "source_keys.user_id"],
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoints_path)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def write_messages(
    spark: SparkSession,
    silver_table: str,
    gold_table: str,
    texts: DataFrame,
    checkpoints_path: str,
) -> StreamingQuery:
    streaming = spark.readStream.format("delta").table(silver_table)

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(spark, gold_table)

        canonical = earliest_version(batch_df, fn.col("source_keys.id"))

        canonical_with_text = canonical.join(texts, "text_version", "left_outer")

        return merge_earliest_version(
            delta_table=delta_table,
            updates=canonical_with_text,
            merge_keys=["source_keys.id"],
        )

    return (
        # filter out auto-generated discord messages types
        streaming.where(
            fn.col("type").isin(arguments.discord_user_message_content_types)
        )
        .writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoints_path)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def write_users(
    spark: SparkSession,
    silver_table: str,
    gold_table: str,
    checkpoints_path: str,
) -> StreamingQuery:
    streaming = spark.readStream.format("delta").table(silver_table)

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(spark, gold_table)

        canonical = most_recent_version(batch_df, fn.col("source_keys.id"))

        return merge_most_recent_version(
            delta_table=delta_table, updates=canonical, merge_keys=["source_keys.id"]
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoints_path)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def write_reactions(
    spark: SparkSession,
    silver_table: str,
    gold_table: str,
    checkpoints_path: str,
) -> StreamingQuery:
    streaming = spark.readStream.format("delta").table(silver_table)

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(spark, gold_table)

        canonical = most_recent_version(batch_df, fn.col("source_keys"))

        return merge_most_recent_version(
            delta_table=delta_table,
            updates=canonical,
            merge_keys=["source_keys.message_id", "source_keys.user_id"],
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoints_path)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    # Guilds
    parser.add_argument(
        "--guilds-silver-table",
        dest="guilds_silver_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--guilds-gold-table",
        dest="guilds_gold_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--guilds-checkpoints-path",
        dest="guilds_checkpoints_path",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    # Channels
    parser.add_argument(
        "--channels-silver-table",
        dest="channels_silver_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--channels-gold-table",
        dest="channels_gold_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--channels-checkpoints-path",
        dest="channels_checkpoints_path",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    # Threads
    parser.add_argument(
        "--threads-silver-table",
        dest="threads_silver_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--threads-gold-table",
        dest="threads_gold_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--threads-checkpoints-path",
        dest="threads_checkpoints_path",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    # Members
    parser.add_argument(
        "--members-silver-table",
        dest="members_silver_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--members-gold-table",
        dest="members_gold_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--members-checkpoints-path",
        dest="members_checkpoints_path",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    # Messages
    parser.add_argument(
        "--messages-silver-table",
        dest="messages_silver_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--texts-silver-table",
        dest="texts_silver_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--messages-gold-table",
        dest="messages_gold_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--messages-checkpoints-path",
        dest="messages_checkpoints_path",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    # Users
    parser.add_argument(
        "--users-silver-table",
        dest="users_silver_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--users-gold-table",
        dest="users_gold_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--users-checkpoints-path",
        dest="users_checkpoints_path",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    # Reactions
    parser.add_argument(
        "--reactions-silver-table",
        dest="reactions_silver_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--reactions-gold-table",
        dest="reactions_gold_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--reactions-checkpoints-path",
        dest="reactions_checkpoints_path",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    write_guilds(
        spark,
        args.guilds_silver_table,
        args.guilds_gold_table,
        args.guilds_checkpoints_path,
    ).awaitTermination()

    write_channels(
        spark,
        args.channels_silver_table,
        args.channels_gold_table,
        args.channels_checkpoints_path,
    ).awaitTermination()

    write_threads(
        spark,
        args.threads_silver_table,
        args.threads_gold_table,
        args.threads_checkpoints_path,
    ).awaitTermination()

    write_members(
        spark,
        args.members_silver_table,
        args.members_gold_table,
        args.members_checkpoints_path,
    ).awaitTermination()

    texts = spark.read.format("delta").table(args.texts_silver_table)

    write_messages(
        spark,
        args.messages_silver_table,
        args.messages_gold_table,
        texts,
        args.messages_checkpoints_path,
    ).awaitTermination()

    write_users(
        spark,
        args.users_silver_table,
        args.users_gold_table,
        args.users_checkpoints_path,
    ).awaitTermination()

    write_reactions(
        spark,
        args.reactions_silver_table,
        args.reactions_gold_table,
        args.reactions_checkpoints_path,
    ).awaitTermination()


if __name__ == "__main__":
    main()
