import argparse
from typing import cast

from delta.tables import DeltaTable
from pyspark.sql import SparkSession, DataFrame
import pyspark.sql.functions as fn
from pyspark.sql.streaming import StreamingQuery

from trend_etl.functions.delta import (
    most_recent_version,
    earliest_version,
    merge_most_recent_version,
    merge_earliest_version,
)

program = "canonicalize_facebook"


def write_groups(
    spark: SparkSession,
    silver_table: str,
    gold_table: str,
    checkpoints_path: str,
) -> StreamingQuery:
    streaming = spark.readStream.format("delta").table(silver_table)

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(spark, gold_table)

        canonical = most_recent_version(batch_df, fn.col("source_keys.id"))

        return merge_most_recent_version(
            delta_table=delta_table, updates=canonical, merge_keys=["source_keys.id"]
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoints_path)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def write_posts(
    spark: SparkSession,
    silver_table: str,
    gold_table: str,
    texts: DataFrame,
    checkpoints_path: str,
) -> StreamingQuery:
    streaming = spark.readStream.format("delta").table(silver_table)

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(spark, gold_table)

        canonical = earliest_version(batch_df, fn.col("source_keys.id"))

        canonical_with_text = canonical.join(texts, "text_version", "left_outer")

        return merge_earliest_version(
            delta_table=delta_table,
            updates=canonical_with_text,
            merge_keys=["source_keys.id"],
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoints_path)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def write_comments(
    spark: SparkSession,
    silver_table: str,
    gold_table: str,
    texts: DataFrame,
    checkpoints_path: str,
) -> StreamingQuery:
    streaming = spark.readStream.format("delta").table(silver_table)

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(spark, gold_table)

        canonical = earliest_version(batch_df, fn.col("source_keys.id"))

        canonical_with_text = canonical.join(texts, "text_version", "left_outer")

        return merge_earliest_version(
            delta_table=delta_table,
            updates=canonical_with_text,
            merge_keys=["source_keys.id"],
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoints_path)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    # Groups
    parser.add_argument(
        "--groups-silver-table",
        dest="groups_silver_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--groups-gold-table",
        dest="groups_gold_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--groups-checkpoints-path",
        dest="groups_checkpoints_path",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    # Posts
    parser.add_argument(
        "--posts-silver-table",
        dest="posts_silver_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--posts-gold-table",
        dest="posts_gold_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--posts-checkpoints-path",
        dest="posts_checkpoints_path",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    # Comments
    parser.add_argument(
        "--comments-silver-table",
        dest="comments_silver_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--comments-gold-table",
        dest="comments_gold_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--comments-checkpoints-path",
        dest="comments_checkpoints_path",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    # Texts
    parser.add_argument(
        "--texts-silver-table",
        dest="texts_silver_table",
        required=True,
        help="The unity catalog table to use",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    write_groups(
        spark,
        args.groups_silver_table,
        args.groups_gold_table,
        args.groups_checkpoints_path,
    ).awaitTermination()

    texts = spark.read.format("delta").table(args.texts_silver_table)

    write_posts(
        spark,
        args.posts_silver_table,
        args.posts_gold_table,
        texts,
        args.posts_checkpoints_path,
    ).awaitTermination()

    write_comments(
        spark,
        args.comments_silver_table,
        args.comments_gold_table,
        texts,
        args.comments_checkpoints_path,
    ).awaitTermination()


if __name__ == "__main__":
    main()
