import argparse
from typing import cast

from delta.tables import DeltaTable
from pyspark.sql import SparkSession, DataFrame
import pyspark.sql.functions as fn
from pyspark.sql.streaming import StreamingQuery

from trend_etl.functions.delta import (
    most_recent_version,
    merge_most_recent_version,
    get_canonical_reddit,
    merge_canonical_reddit,
)

program = "canonicalize_reddit"


def write_subreddits(
    spark: SparkSession,
    silver_table: str,
    gold_table: str,
    checkpoints_path: str,
) -> StreamingQuery:
    streaming = spark.readStream.format("delta").table(silver_table)

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(spark, gold_table)

        canonical = most_recent_version(batch_df, fn.col("source_keys.name"))

        return merge_most_recent_version(
            delta_table=delta_table, updates=canonical, merge_keys=["source_keys.name"]
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoints_path)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def write_submissions(
    spark: SparkSession,
    silver_table: str,
    gold_table: str,
    texts: DataFrame,
    subreddit_idents: DataFrame,
    checkpoints_path: str,
) -> StreamingQuery:
    streaming = spark.readStream.format("delta").table(silver_table)

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(spark, gold_table)

        # NOTE: filter out some submissions that may not link to the subreddits table
        # (this affects a small amount of data from pushshift)
        # See "Missing Subreddits" section https://wiki.trend.community/en/technology/data-lake
        filtered = batch_df.join(
            subreddit_idents,
            fn.col("source_keys.subreddit_id") == fn.col("id"),
            "left_semi",
        )

        canonical = get_canonical_reddit(filtered, "submission")

        canonical_with_text = canonical.join(texts, "text_version", "left_outer")

        return merge_canonical_reddit(
            delta_table=delta_table, updates=canonical_with_text, schema="submission"
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoints_path)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def write_replies(
    spark: SparkSession,
    silver_table: str,
    gold_table: str,
    texts: DataFrame,
    subreddit_idents: DataFrame,
    checkpoints_path: str,
) -> StreamingQuery:
    streaming = spark.readStream.format("delta").table(silver_table)

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(spark, gold_table)

        # NOTE: filter out some submissions that may not link to the subreddits table
        # (this affects a small amount of data from pushshift)
        # See "Missing Subreddits" section https://wiki.trend.community/en/technology/data-lake
        filtered = batch_df.join(
            subreddit_idents,
            fn.col("source_keys.subreddit_id") == fn.col("id"),
            "left_semi",
        )

        canonical = get_canonical_reddit(filtered, "reply")

        canonical_with_text = canonical.join(texts, "text_version", "left_outer")

        return merge_canonical_reddit(
            delta_table=delta_table, updates=canonical_with_text, schema="reply"
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoints_path)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    # Subreddits
    parser.add_argument(
        "--subreddits-silver-table",
        dest="subreddits_silver_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--subreddits-gold-table",
        dest="subreddits_gold_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--subreddits-checkpoints-path",
        dest="subreddits_checkpoints_path",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    # Submissions
    parser.add_argument(
        "--submissions-silver-table",
        dest="submissions_silver_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--submissions-gold-table",
        dest="submissions_gold_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--submissions-checkpoints-path",
        dest="submissions_checkpoints_path",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    # Replies
    parser.add_argument(
        "--replies-silver-table",
        dest="replies_silver_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--replies-gold-table",
        dest="replies_gold_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--replies-checkpoints-path",
        dest="replies_checkpoints_path",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    # Subreddit idents
    parser.add_argument(
        "--subreddit-idents-table",
        dest="subreddit_idents_table",
        required=True,
        help="The unity catalog table to use",
    )

    # Texts
    parser.add_argument(
        "--texts-silver-table",
        dest="texts_silver_table",
        required=True,
        help="The unity catalog table to use",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    subreddit_idents = spark.read.format("delta").table(args.subreddit_idents_table)

    texts = spark.read.format("delta").table(args.texts_silver_table)

    write_subreddits(
        spark,
        args.subreddits_silver_table,
        args.subreddits_gold_table,
        args.subreddits_checkpoints_path,
    ).awaitTermination()

    write_submissions(
        spark,
        args.submissions_silver_table,
        args.submissions_gold_table,
        texts,
        subreddit_idents,
        args.submissions_checkpoints_path,
    ).awaitTermination()

    write_replies(
        spark,
        args.replies_silver_table,
        args.replies_gold_table,
        texts,
        subreddit_idents,
        args.replies_checkpoints_path,
    ).awaitTermination()


if __name__ == "__main__":
    main()
