import argparse
from typing import cast

# pylint: disable-next=import-error,no-name-in-module
from pyspark.dbutils import DBUtils  # type: ignore
from pyspark.sql import SparkSession
import pyspark.sql.functions as fn

# pylint: disable-next=import-error,no-name-in-module
import psycopg2  # type: ignore

# pylint: disable-next=import-error,no-name-in-module
from datasets import Dataset, Features, Value  # type: ignore

# pylint: disable-next=import-error,no-name-in-module
from transformers import AutoTokenizer, DataCollatorWithPadding, TFAutoModelForSequenceClassification  # type: ignore

# pylint: disable-next=import-error,no-name-in-module
import tensorflow as tf  # type: ignore

import trend_etl.schema.args as arguments

program = "classify_new_subreddits"
postgres_host = "trend-prodogy-instance.cs7lhsrxph4x.us-east-1.rds.amazonaws.com"
postgres_db = "datapipelines"
new_subreddits_table = "prod.new_subreddits"
postgres_port = 5432
postgres_user = "databricks"
model_path = "/dbfs/FileStore/subreddit_expansion/models/distilbert-uncased-8/"
threshold_prob = (
    0.876  # subreddits above this probability was giving good results with this model
)


def get_postgres_password(spark: SparkSession) -> str:
    dbutils = DBUtils(spark)
    token = dbutils.secrets.get(scope="postgres", key="password")

    return token


# Below code can be used to save and load tokenizer:
#
# from transformers import AutoTokenizer, AutoConfig
# tokenizer = AutoTokenizer.from_pretrained('roberta-base')
# tokenizer.save_pretrained('TEST/tokenizer')
# tokenizer = AutoTokenizer.from_pretrained('TEST/tokenizer', config=AutoConfig.from_pretrained("roberta-base"))
def prep_df_for_prediction(df, model):
    """Prepare a dataframe of subreddits for prediction.
    `df` is assumed to be a pandas dataframe with 'subreddit'
    as the index, and with (string) columns for:
    - title
    - advertiser_category
    - public_description
    - description
    - submit_text
    `model` is the model the dataframe is being prepared for."""
    # Create a Dataset object from the dataframe
    features = Features({"text": Value("string")})
    subs = Dataset.from_spark(
        df, features=features, cache_dir="/dbfs/FileStore/classify_subreddits_model"
    )

    # Tokenize the input text
    tokenizer = AutoTokenizer.from_pretrained("distilbert-base-uncased")

    def preprocess(examples):
        return tokenizer(examples["text"], truncation=True)

    tokenized_subs = subs.map(preprocess, batched=True)

    data_collator = DataCollatorWithPadding(tokenizer=tokenizer, return_tensors="tf")
    return model.prepare_tf_dataset(
        tokenized_subs, shuffle=False, batch_size=16, collate_fn=data_collator
    )


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    parser.add_argument(
        "--target-catalog",
        dest="target_catalog",
        required=True,
        choices=[
            arguments.catalogs[arguments.catalogs_dev],
            arguments.catalogs[arguments.catalogs_prod],
        ],
        help="The unit catalog to use",
    )
    args = parser.parse_args()

    if args.target_catalog == arguments.catalogs[arguments.catalogs_dev]:
        return  # do not run dev version of this job

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())
    postgres_password = get_postgres_password(spark)

    new_subreddits = (
        spark.read.format("postgresql")
        .option("dbtable", new_subreddits_table)
        .option("host", postgres_host)
        .option("port", postgres_port)
        .option("database", postgres_db)
        .option("user", postgres_user)
        .option("password", postgres_password)
        .load()
        .where(fn.col("classification_status") == "new")
        .selectExpr("id", "data")
        .cache()
    )

    num_new = new_subreddits.count()
    print("New subreddits", num_new)

    if num_new > 0:
        new_subreddits_text = new_subreddits.select(
            fn.col("id"),
            fn.format_string(
                "<name> %s <title> %s <ad_cat> %s <pub_desc> %s <description> %s <sub> %s",
                fn.coalesce(
                    fn.get_json_object("data", "$.data.display_name"), fn.lit("")
                ),
                fn.coalesce(fn.get_json_object("data", "$.data.title"), fn.lit("")),
                fn.coalesce(
                    fn.get_json_object("data", "$.data.advertiser_category"), fn.lit("")
                ),
                fn.coalesce(
                    fn.get_json_object("data", "$.data.public_description"), fn.lit("")
                ),
                fn.coalesce(
                    fn.get_json_object("data", "$.data.description"), fn.lit("")
                ),
                fn.coalesce(
                    fn.get_json_object("data", "$.data.submit_text"), fn.lit("")
                ),
            ).alias("text"),
        )

        mymodel = TFAutoModelForSequenceClassification.from_pretrained(model_path)

        data = prep_df_for_prediction(
            new_subreddits_text.select(fn.col("text")), mymodel
        )

        logits = mymodel.predict(data).logits
        probs = tf.nn.softmax(logits)[:, 1]

        pandas = new_subreddits_text.select(fn.col("id")).toPandas()
        pandas["prob"] = probs

        result_df = spark.createDataFrame(pandas)

        updates = result_df.withColumn(
            "classification_status",
            fn.when((fn.col("prob") > threshold_prob), fn.lit("add")).otherwise(
                fn.lit("ignore")
            ),
        ).toPandas()

        conn = psycopg2.connect(
            dbname=postgres_db,
            user=postgres_user,
            password=postgres_password,
            host=postgres_host,
            port=postgres_port,
        )

        # Disable autocommit mode (should be default)
        conn.autocommit = False

        try:
            with conn:
                with conn.cursor() as cur:
                    for _index, row in updates.iterrows():
                        # Create the SQL query to update the records
                        update_query = """
                            update prod.new_subreddits
                            set classification_status = %s,
                                prob = %s
                            where id = %s
                        """

                        cur.execute(
                            update_query,
                            (row["classification_status"], row["prob"], row["id"]),
                        )
        finally:
            conn.close()


if __name__ == "__main__":
    main()
