import argparse
from typing import cast

# pylint: disable-next=import-error,no-name-in-module
from pyspark.dbutils import DBUtils  # type: ignore
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.functions import to_json


program = "dump_discord_metadata"


def get_postgres_password(spark: SparkSession) -> str:
    dbutils = DBUtils(spark)

    token = dbutils.secrets.get(scope="postgres", key="password")

    return token


def overwrite_pg_table(
    df: DataFrame,
    pg_host: str,
    pg_port: str,
    pg_db: str,
    pg_user: str,
    pg_password: str,
    pg_table: str,
) -> None:
    (
        df.write.format("postgresql")
        .mode("overwrite")
        .option("truncate", "true")
        .option("dbtable", pg_table)
        .option("host", pg_host)
        .option("port", pg_port)
        .option("database", pg_db)
        .option("user", pg_user)
        .option("password", pg_password)
        .save()
    )


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    # Postgres
    parser.add_argument(
        "--pg-host",
        dest="pg_host",
        required=True,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--pg-port",
        dest="pg_port",
        required=True,
        type=int,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--pg-db",
        dest="pg_db",
        required=True,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--pg-user",
        dest="pg_user",
        required=True,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--guilds-pg-table",
        dest="guilds_pg_table",
        required=True,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--channels-pg-table",
        dest="channels_pg_table",
        required=True,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--messages-pg-table",
        dest="messages_pg_table",
        required=True,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--reactions-pg-table",
        dest="reactions_pg_table",
        required=True,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--members-pg-table",
        dest="members_pg_table",
        required=True,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--users-pg-table",
        dest="users_pg_table",
        required=True,
        help="Postgres parameter",
    )

    # Tables
    parser.add_argument(
        "--guilds-table",
        dest="guilds_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--channels-table",
        dest="channels_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--messages-table",
        dest="messages_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--reactions-table",
        dest="reactions_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--members-table",
        dest="members_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--users-table",
        dest="users_table",
        required=True,
        help="The unity catalog table to use",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    pg_password = get_postgres_password(spark)

    # Guilds
    guilds = spark.table(args.guilds_table).withColumn(
        "source_keys", to_json("source_keys")
    )

    overwrite_pg_table(
        guilds,
        args.pg_host,
        args.pg_port,
        args.pg_db,
        args.pg_user,
        pg_password,
        args.guilds_pg_table,
    )

    # Channels
    channels = (
        spark.table(args.channels_table)
        .withColumn("source_keys", to_json("source_keys"))
        .withColumn("permission_overwrites", to_json("permission_overwrites"))
        .withColumn("thread_metadata", to_json("thread_metadata"))
        .withColumn("available_tags", to_json("available_tags"))
        .withColumn("default_reaction_emoji", to_json("default_reaction_emoji"))
    )

    overwrite_pg_table(
        channels,
        args.pg_host,
        args.pg_port,
        args.pg_db,
        args.pg_user,
        pg_password,
        args.channels_pg_table,
    )

    # Messages
    messages = (
        spark.table(args.messages_table)
        .drop("text")
        .withColumn("source_keys", to_json("source_keys"))
        .withColumn("embeds", to_json("embeds"))
        .withColumn("thread", to_json("thread"))
        .withColumn("sticker_items", to_json("sticker_items"))
        .withColumn("poll", to_json("poll"))
        .withColumn("call", to_json("call"))
    )

    overwrite_pg_table(
        messages,
        args.pg_host,
        args.pg_port,
        args.pg_db,
        args.pg_user,
        pg_password,
        args.messages_pg_table,
    )

    # Reactions
    reactions = (
        spark.table(args.reactions_table)
        .withColumn("source_keys", to_json("source_keys"))
        .withColumn("emoji", to_json("emoji"))
    )

    overwrite_pg_table(
        reactions,
        args.pg_host,
        args.pg_port,
        args.pg_db,
        args.pg_user,
        pg_password,
        args.reactions_pg_table,
    )

    # Members
    members = (
        spark.table(args.members_table)
        .withColumn("source_keys", to_json("source_keys"))
        .withColumn("avatar_decoration_data", to_json("avatar_decoration_data"))
    )

    overwrite_pg_table(
        members,
        args.pg_host,
        args.pg_port,
        args.pg_db,
        args.pg_user,
        pg_password,
        args.members_pg_table,
    )

    # Users
    users = spark.table(args.users_table).withColumn(
        "source_keys", to_json("source_keys")
    )

    overwrite_pg_table(
        users,
        args.pg_host,
        args.pg_port,
        args.pg_db,
        args.pg_user,
        pg_password,
        args.users_pg_table,
    )


if __name__ == "__main__":
    main()
