import argparse
from typing import cast

from delta.tables import DeltaTable
from pyspark.sql import SparkSession

from trend_etl.functions.delta import (
    community_group_upsert_to_delta,
)
from trend_etl.transform.community_groups import from_bronze_get_community_groups_df

program = "load_community_groups"


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    parser.add_argument(
        "--idents-table",
        dest="idents_table",
        required=True,
        help="The unity catalog table to use",
    )

    parser.add_argument(
        "--dst-table",
        dest="dst_table",
        required=True,
        help="The unity catalog table to use",
    )

    parser.add_argument(
        "--data-source",
        dest="data_source",
        required=True,
        help="The social media data source",
    )

    parser.add_argument(
        "--checkpoints-path",
        dest="checkpoints_path",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    streaming = spark.readStream.format("delta").table(args.idents_table)

    prepared = from_bronze_get_community_groups_df(streaming, args.data_source)

    delta_table = DeltaTable.forName(spark, args.dst_table)

    (
        prepared.writeStream.format("delta")
        .foreachBatch(
            lambda batch_df, batch_id: community_group_upsert_to_delta(
                delta_table, batch_df
            )
        )
        .option("checkpointLocation", args.checkpoints_path)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


if __name__ == "__main__":
    main()
