"""
Load docs table for new documents fetched from facebook, reddit, discord, etc.

NOTE: All load_docs_* functions are safe to re-run (idempotent).
"""

import argparse
from typing import cast

from delta.tables import DeltaTable
from pyspark.sql import SparkSession, DataFrame
import pyspark.sql.functions as fn
from pyspark.sql.streaming import StreamingQuery

from trend_etl.functions.delta import (
    full_table_name,
    merge_earliest_collected_by_key,
)
from trend_etl.transform.posts import from_bronze_get_posts_df
from trend_etl.transform.comments import from_bronze_get_comments_df
from trend_etl.transform.submissions import from_bronze_get_submissions_df_append
from trend_etl.transform.replies import from_bronze_get_replies_df_append
from trend_etl.transform.messages import from_bronze_get_messages_df
import trend_etl.schema.args as arguments
import trend_etl.schema.fields as fields

PROGRAM = "load_docs_append"


def load_docs_posts(
    spark: SparkSession,
    bronze_catalog: str,
    docs_table: DeltaTable,
    group_idents: DataFrame,
    post_idents: DataFrame,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    """
    Merge in new documents added to the submissions bronze table.
    """
    bronze_table = full_table_name(
        bronze_catalog,
        arguments.schemas[arguments.schemas_bronze],
        arguments.tables[arguments.tables_posts],
    )
    streaming = spark.readStream.format("delta").table(bronze_table)
    checkpoint_loc = (
        checkpoints_path_prefix
        + "/docs_append/"
        + arguments.tables[arguments.tables_posts]
    )

    texts_col = fn.struct(
        fn.col(fields.facebook_posts_title).alias("title"),
        fn.col(fields.facebook_posts_text).alias("text"),
    )
    prepared = from_bronze_get_posts_df(streaming, group_idents, post_idents).select(
        fn.col(fields.idents_doc_id_field),
        fn.col(fields.idents_group_id_field),
        fn.col(fields.source_created_at),
        texts_col.alias(fields.docs_texts),
        fn.col(fields.bronze_timestamp_field),
    )

    return (
        prepared.writeStream.format("delta")
        .foreachBatch(
            lambda batch_df, batch_id: merge_earliest_collected_by_key(
                docs_table, batch_df, [fields.idents_doc_id_field]
            )
        )
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def load_docs_comments(
    spark: SparkSession,
    bronze_catalog: str,
    docs_table: DeltaTable,
    group_idents: DataFrame,
    post_idents: DataFrame,
    comment_idents: DataFrame,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    """
    Merge in new documents added to the submissions bronze table.
    """
    bronze_table = full_table_name(
        bronze_catalog,
        arguments.schemas[arguments.schemas_bronze],
        arguments.tables[arguments.tables_comments],
    )
    streaming = spark.readStream.format("delta").table(bronze_table)
    checkpoint_loc = (
        checkpoints_path_prefix
        + "/docs_append/"
        + arguments.tables[arguments.tables_comments]
    )

    texts_col = fn.struct(
        fn.expr("cast(null as string)").alias("title"),
        fn.col(fields.facebook_comments_text).alias("text"),
    )
    prepared = from_bronze_get_comments_df(
        streaming, group_idents, post_idents, comment_idents
    ).select(
        fn.col(fields.idents_doc_id_field),
        fn.col(fields.idents_group_id_field),
        fn.col(fields.source_created_at),
        texts_col.alias(fields.docs_texts),
        fn.col(fields.bronze_timestamp_field),
    )

    return (
        prepared.writeStream.format("delta")
        .foreachBatch(
            lambda batch_df, batch_id: merge_earliest_collected_by_key(
                docs_table, batch_df, [fields.idents_doc_id_field]
            )
        )
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def load_docs_submissions(
    spark: SparkSession,
    bronze_catalog: str,
    docs_table: DeltaTable,
    subreddit_idents: DataFrame,
    submission_idents: DataFrame,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    """
    Merge in new documents added to the submissions bronze table.
    """
    bronze_table = full_table_name(
        bronze_catalog,
        arguments.schemas[arguments.schemas_bronze],
        arguments.tables[arguments.tables_submissions],
    )
    streaming = spark.readStream.format("delta").table(bronze_table)
    checkpoint_loc = (
        checkpoints_path_prefix
        + "/docs_append/"
        + arguments.tables[arguments.tables_submissions]
    )

    texts_col = fn.struct(
        fn.col(fields.reddit_submissions_title).alias("title"),
        fn.col(fields.reddit_submissions_text).alias("text"),
    )
    prepared = from_bronze_get_submissions_df_append(
        streaming, subreddit_idents, submission_idents
    ).select(
        fn.col(fields.idents_doc_id_field),
        fn.col(fields.idents_group_id_field),
        fn.col(fields.source_created_at),
        texts_col.alias(fields.docs_texts),
        fn.col(fields.bronze_timestamp_field),
    )

    return (
        prepared.writeStream.format("delta")
        .foreachBatch(
            lambda batch_df, batch_id: merge_earliest_collected_by_key(
                docs_table, batch_df, [fields.idents_doc_id_field]
            )
        )
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def load_docs_replies(
    spark: SparkSession,
    bronze_catalog: str,
    docs_table: DeltaTable,
    subreddit_idents: DataFrame,
    submission_idents: DataFrame,
    reply_idents: DataFrame,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    """
    Merge in new documents added to the replies bronze table.
    """
    bronze_table = full_table_name(
        bronze_catalog,
        arguments.schemas[arguments.schemas_bronze],
        arguments.tables[arguments.tables_replies],
    )
    streaming = spark.readStream.format("delta").table(bronze_table)
    checkpoint_loc = (
        checkpoints_path_prefix
        + "/docs_append/"
        + arguments.tables[arguments.tables_replies]
    )

    texts_col = fn.struct(
        fn.expr("cast(null as string)").alias("title"),
        fn.col(fields.reddit_replies_text).alias("text"),
    )
    prepared = from_bronze_get_replies_df_append(
        streaming,
        subreddit_idents,
        submission_idents,
        reply_idents,
    ).select(
        fn.col(fields.idents_doc_id_field),
        fn.col(fields.idents_group_id_field),
        fn.col(fields.source_created_at),
        texts_col.alias(fields.docs_texts),
        fn.col(fields.bronze_timestamp_field),
    )

    return (
        prepared.writeStream.format("delta")
        .foreachBatch(
            lambda batch_df, batch_id: merge_earliest_collected_by_key(
                docs_table, batch_df, [fields.idents_doc_id_field]
            )
        )
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def load_docs_discord(
    spark: SparkSession,
    bronze_catalog: str,
    docs_table: DeltaTable,
    guild_idents: DataFrame,
    channel_idents: DataFrame,
    message_idents: DataFrame,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    """
    Merge in new documents added to the replies bronze table.
    """
    bronze_table = full_table_name(
        bronze_catalog,
        arguments.schemas[arguments.schemas_bronze],
        arguments.tables[arguments.tables_messages],
    )
    streaming = spark.readStream.format("delta").table(bronze_table)
    checkpoint_loc = (
        checkpoints_path_prefix
        + "/docs_append/"
        + arguments.tables[arguments.tables_messages]
    )

    texts_col = fn.struct(
        fn.expr("cast(null as string)").alias("title"),
        fn.col("content").alias("text"),
    )
    prepared = from_bronze_get_messages_df(
        streaming,
        guild_idents,
        channel_idents,
        message_idents,
    ).select(
        fn.col(fields.idents_doc_id_field),
        fn.col(fields.idents_group_id_field),
        fn.col(fields.source_created_at),
        texts_col.alias(fields.docs_texts),
        fn.col(fields.bronze_timestamp_field),
    )

    return (
        prepared.writeStream.format("delta")
        .foreachBatch(
            lambda batch_df, batch_id: merge_earliest_collected_by_key(
                docs_table, batch_df, [fields.idents_doc_id_field]
            )
        )
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def main() -> None:
    """
    Entry point for load docs task
    """
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{PROGRAM}.")

    parser.add_argument(
        "--bronze-catalog",
        dest="bronze_catalog",
        required=True,
        choices=arguments.catalogs,
        help="The unit catalog to use",
    )

    parser.add_argument(
        "--target-catalog",
        dest="target_catalog",
        required=True,
        choices=arguments.catalogs,
        help="The unit catalog to use",
    )

    parser.add_argument(
        "--checkpoints-path-prefix",
        dest="checkpoints_path_prefix",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(PROGRAM).getOrCreate())

    subreddit_idents_table = full_table_name(
        args.target_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_subreddit_idents],
    )
    subreddit_idents = spark.read.format("delta").table(subreddit_idents_table)

    submission_idents_table = full_table_name(
        args.target_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_submission_idents],
    )
    submission_idents = spark.read.format("delta").table(submission_idents_table)

    reply_idents_table = full_table_name(
        args.target_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_reply_idents],
    )
    reply_idents = spark.read.format("delta").table(reply_idents_table)

    guild_idents_table = full_table_name(
        args.target_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_guild_idents],
    )
    guild_idents = spark.read.format("delta").table(guild_idents_table)

    channel_idents_table = full_table_name(
        args.target_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_channel_idents],
    )
    channel_idents = spark.read.format("delta").table(channel_idents_table)

    message_idents_table = full_table_name(
        args.target_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_message_idents],
    )
    message_idents = spark.read.format("delta").table(message_idents_table)

    group_idents_table = full_table_name(
        args.target_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_group_idents],
    )
    group_idents = spark.read.format("delta").table(group_idents_table)

    post_idents_table = full_table_name(
        args.target_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_post_idents],
    )
    post_idents = spark.read.format("delta").table(post_idents_table)

    comment_idents_table = full_table_name(
        args.target_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_comment_idents],
    )
    comment_idents = spark.read.format("delta").table(comment_idents_table)

    tablename = full_table_name(
        args.target_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_docs_append],
    )
    docs_table = DeltaTable.forName(spark, tablename)

    load_docs_posts(
        spark,
        args.bronze_catalog,
        docs_table,
        group_idents,
        post_idents,
        args.checkpoints_path_prefix,
    ).awaitTermination()

    load_docs_comments(
        spark,
        args.bronze_catalog,
        docs_table,
        group_idents,
        post_idents,
        comment_idents,
        args.checkpoints_path_prefix,
    ).awaitTermination()

    load_docs_submissions(
        spark,
        args.bronze_catalog,
        docs_table,
        subreddit_idents,
        submission_idents,
        args.checkpoints_path_prefix,
    ).awaitTermination()

    load_docs_replies(
        spark,
        args.bronze_catalog,
        docs_table,
        subreddit_idents,
        submission_idents,
        reply_idents,
        args.checkpoints_path_prefix,
    ).awaitTermination()

    load_docs_discord(
        spark,
        args.bronze_catalog,
        docs_table,
        guild_idents,
        channel_idents,
        message_idents,
        args.checkpoints_path_prefix,
    ).awaitTermination()


if __name__ == "__main__":
    main()
