import argparse
from typing import cast

from delta.tables import DeltaTable
from pyspark.sql import SparkSession
from pyspark.sql.streaming import StreamingQuery

from trend_etl.functions.delta import (
    ao_merge_using_ingestion_key,
    ao_merge_using_text_key,
)
from trend_etl.transform.submissions import (
    from_bronze_get_submissions_df_ao,
    from_bronze_get_submissions_text_df_ao,
)

program = "load_submissions_ao"


def load_submissions(
    spark: SparkSession,
    bronze_table: str,
    idents_table: str,
    dst_table: str,
    checkpoints_path: str,
) -> StreamingQuery:
    streaming = spark.readStream.format("delta").table(bronze_table)

    idents = spark.read.format("delta").table(idents_table)

    prepared = from_bronze_get_submissions_df_ao(streaming, idents)

    table = DeltaTable.forName(spark, dst_table)

    return (
        prepared.writeStream.format("delta")
        .foreachBatch(
            lambda batch_df, batch_id: ao_merge_using_ingestion_key(table, batch_df)
        )
        .option("checkpointLocation", checkpoints_path)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def load_texts(
    spark: SparkSession,
    bronze_table: str,
    dst_table: str,
    checkpoints_path: str,
) -> StreamingQuery:
    streaming = spark.readStream.format("delta").table(bronze_table)

    prepared = from_bronze_get_submissions_text_df_ao(streaming)

    table = DeltaTable.forName(spark, dst_table)

    return (
        prepared.writeStream.format("delta")
        .foreachBatch(
            lambda batch_df, batch_id: ao_merge_using_text_key(table, batch_df)
        )
        .option("checkpointLocation", checkpoints_path)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    parser.add_argument(
        "--bronze-table",
        dest="bronze_table",
        required=True,
        help="The unity catalog table to use",
    )

    parser.add_argument(
        "--idents-table",
        dest="idents_table",
        required=True,
        help="The unity catalog table to use",
    )

    parser.add_argument(
        "--dst-table",
        dest="dst_table",
        required=True,
        help="The unity catalog table to use",
    )

    parser.add_argument(
        "--texts-table",
        dest="texts_table",
        required=True,
        help="The unity catalog table to use",
    )

    parser.add_argument(
        "--checkpoints-path",
        dest="checkpoints_path",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    parser.add_argument(
        "--texts-checkpoints-path",
        dest="texts_checkpoints_path",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    load_submissions(
        spark,
        args.bronze_table,
        args.idents_table,
        args.dst_table,
        args.checkpoints_path,
    ).awaitTermination()

    load_texts(
        spark,
        args.bronze_table,
        args.texts_table,
        args.texts_checkpoints_path,
    ).awaitTermination()


if __name__ == "__main__":
    main()
