import argparse
from typing import cast, Generator
import json

from slack_sdk import WebClient

# pylint: disable-next=import-error,no-name-in-module
from pyspark.dbutils import DBUtils  # type: ignore
from pyspark.sql import SparkSession, DataFrame
from trend_etl.transform.discord_stats import (
    total_registered_users_number,
    total_journeys_number,
    journey_conversion_number,
    total_statements_number,
    active_users_days_ago_number,
    active_audience_days_ago_number,
    retention_audience_number,
)

program = "report_discord_stats"
message_count_threshold = 4
reaction_count_threshold = 2


def get_slack_api_token(spark: SparkSession) -> str:
    dbutils = DBUtils(spark)

    token = dbutils.secrets.get(scope="slack", key="trendy")

    return token


def report_registered_users(
    users: DataFrame,
):
    registered_users = total_registered_users_number(users)
    message = f"{registered_users} Total Registered Users"

    return message


def report_total_statements(
    threads: DataFrame,
    messages: DataFrame,
):
    total_statements = total_statements_number(threads, messages)
    message = f"{total_statements} Total Statements"

    return message


def report_total_journeys(
    channels: DataFrame,
    users: DataFrame,
    journey_parent_channels: list[str],
):
    total_journeys = total_journeys_number(channels, users, journey_parent_channels)
    message = (
        f"{total_journeys} Journeys "
        f"(over {len(journey_parent_channels)} channel{'s' if len(journey_parent_channels) > 1 else ''})"
    )

    return message


def report_journey_conversion(
    channels: DataFrame,
    users: DataFrame,
    journey_parent_channels: list[str],
):
    journeys_conversion = journey_conversion_number(
        channels, users, journey_parent_channels
    )
    message = f"Journey Conversion: {journeys_conversion} users created journey(s)"

    return message


def report_active_users_stats(
    threads: DataFrame,
    messages: DataFrame,
) -> Generator[str, None, None]:
    for days_ago in (30, 60, 90):
        active_users = active_users_days_ago_number(
            threads, messages, days_ago, message_count_threshold
        )
        message = f"{active_users} active users in the last {days_ago} days"

        yield message


def report_active_audience_stats(
    threads: DataFrame,
    messages: DataFrame,
    reactions: DataFrame,
    users: DataFrame,
) -> Generator[str, None, None]:
    for days_ago in (30,):
        active_audience = active_audience_days_ago_number(
            threads,
            messages,
            reactions,
            users,
            days_ago,
            message_count_threshold,
            reaction_count_threshold,
        )
        message = f"{active_audience} active audience in the last {days_ago} days"

        yield message


# NOTE: Not used at the moment
def report_audience_retention(
    threads: DataFrame,
    messages: DataFrame,
    reactions: DataFrame,
    users: DataFrame,
    days_partition: int,
):
    audience_retention = retention_audience_number(
        threads, messages, reactions, users, days_partition
    )
    message = (
        f"{audience_retention} retained for an active audience "
        f"(in the last {days_partition} days)"
    )

    return message


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    # Discord Tables
    parser.add_argument(
        "--channels-table",
        dest="channels_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--threads-table",
        dest="threads_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--messages-table",
        dest="messages_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--reactions-table",
        dest="reactions_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--users-table",
        dest="users_table",
        required=True,
        help="The unity catalog table to use",
    )

    # JSON schema [string]
    parser.add_argument(
        "--journey-channels",
        dest="journey_channels",
        required=True,
        help="A JSON list of channels (dedicated for journeys) to collect data for",
    )

    parser.add_argument(
        "--slack-channel-id",
        dest="slack_channel_id",
        required=True,
        help="The unique databricks (job) run id that runs this",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    journey_parent_channels = json.loads(args.journey_channels)

    trendy_token = get_slack_api_token(spark)
    _slack_client = WebClient(token=trendy_token)

    channels = spark.read.format("delta").table(args.channels_table)
    threads = spark.read.format("delta").table(args.threads_table)
    messages = spark.read.format("delta").table(args.messages_table)
    reactions = spark.read.format("delta").table(args.reactions_table)
    users = spark.read.format("delta").table(args.users_table)

    message_comps = (
        [
            "Discord Daily Stats",
            report_registered_users(users),
            report_total_statements(threads, messages),
            report_total_journeys(channels, users, journey_parent_channels),
            report_journey_conversion(channels, users, journey_parent_channels),
        ]
        + [
            "\n",
            (
                "Active users are defined as creating "
                f"{message_count_threshold} or more messages "
                "within the below time periods"
            ),
        ]
        + list(report_active_users_stats(threads, messages))
        + [
            "\n",
            (
                "Active audience are active users plus "
                "users that have reacted "
                f"{reaction_count_threshold} or more times"
            ),
        ]
        + list(
            report_active_audience_stats(
                threads,
                messages,
                reactions,
                users,
            )
        )
    )

    # slack_client.chat_postMessage(
    #     channel=args.slack_channel_id,
    #     text="\n".join(message_comps),
    # )

    print("\n".join(message_comps))


if __name__ == "__main__":
    main()
