import argparse
from typing import cast
import json

from slack_sdk import WebClient

from delta.tables import DeltaTable

# pylint: disable-next=import-error,no-name-in-module
from pyspark.dbutils import DBUtils  # type: ignore
from pyspark.sql import SparkSession
import pyspark.sql.functions as fn

program = "report_statements_ao"


def report_message_for_notifier(
    spark: SparkSession,
    _slack_client: WebClient,
    _slack_channel_id: str,
    table_notifier: dict,
) -> None:
    delta_table = DeltaTable.forName(spark, table_notifier["table"])
    history_rows = (
        delta_table.history()
        .orderBy(fn.col("version").desc())
        .selectExpr("version")
        .limit(2)
        .collect()
    )
    assert len(history_rows) == 2
    version_current: int = history_rows[0]["version"]
    version_prior: int = history_rows[1]["version"]

    current_version_df = spark.read.option("versionAsOf", version_current).table(
        table_notifier["table"]
    )

    prior_version_df = (
        spark.read.option("versionAsOf", version_prior)
        .table(table_notifier["table"])
        .cache()
    )

    current_version_delta = current_version_df.join(
        prior_version_df, on="ingest_version", how="left_anti"
    ).cache()

    new_docs = (
        current_version_delta.alias("delta")
        .join(
            prior_version_df.alias("prior"),
            fn.col("delta.source_keys.name") == fn.col("prior.source_keys.name"),
            "left_anti",
        )
        .cache()
    )
    insert_count = (
        new_docs.withColumn("_doc_id", fn.col("source_keys.name"))
        .dropDuplicates(["_doc_id"])
        .count()
    )

    if insert_count == 0:
        message = f"No new {table_notifier['notification_label']} inserted"
    else:
        message = (
            f"{table_notifier['notification_label']} inserted {insert_count} statements"
        )

    # slack_client.chat_postMessage(channel=slack_channel_id, text=message)
    print(message)


def get_slack_api_token(spark: SparkSession) -> str:
    dbutils = DBUtils(spark)

    token = dbutils.secrets.get(scope="slack", key="trendy")

    return token


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    # JSON schema {table:string, notification_label:string}
    parser.add_argument(
        "--table-notifier",
        dest="table_notifier",
        required=True,
        help="A JSON str that is a table notifiers for reporting",
    )

    parser.add_argument(
        "--slack-channel-id",
        dest="slack_channel_id",
        required=True,
        help="The unique databricks (job) run id that runs this",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    table_notifier = json.loads(args.table_notifier)

    trendy_token = get_slack_api_token(spark)
    slack_client = WebClient(token=trendy_token)

    report_message_for_notifier(
        spark, slack_client, args.slack_channel_id, table_notifier
    )


if __name__ == "__main__":
    main()
