import unittest

from pyspark.sql import SparkSession
from pyspark.sql.functions import reverse, col, collect_set, size
from pyspark.sql.types import StringType
from trend_etl.test_utils import (
    getspark,
    get_kafka_df,
    get_archive_df,
    get_df,
    assert_dataframes_equal,
)
from trend_etl.transform.groups import (
    from_kafka_get_bronze_groups_df,
    from_archive_get_bronze_groups_df,
)
from trend_etl.transform.posts import (
    from_kafka_get_bronze_posts_df,
    from_archive_get_bronze_posts_df,
)
from trend_etl.transform.comments import (
    from_kafka_get_bronze_comments_df,
    from_archive_get_bronze_comments_df,
)
from trend_etl.transform.subreddits import (
    from_kafka_get_bronze_subreddits_df,
    from_bronze_get_subreddits_df,
    from_archive_get_bronze_subreddits_df,
)
from trend_etl.transform.submissions import (
    from_kafka_get_bronze_submissions_df,
    from_reddit_archive_get_bronze_submissions_df,
    from_pushshift_archive_get_bronze_submissions_df,
)
from trend_etl.transform.replies import (
    from_kafka_get_bronze_replies_df,
    from_reddit_archive_get_bronze_replies_df,
    from_pushshift_archive_get_bronze_replies_df,
)
from trend_etl.transform.idents import from_bronze_get_idents_df
from trend_etl.transform.guilds import from_kafka_get_bronze_guilds_df
from trend_etl.transform.channels import from_kafka_get_bronze_channels_df
from trend_etl.transform.messages import from_kafka_get_bronze_messages_df
from trend_etl.transform.reactions import from_kafka_get_bronze_reactions_df
from trend_etl.functions.misc import get_edited_flag_dataframe
from trend_etl.functions.parsing import parse_reddit_submissions
import trend_etl.schema.fields as fields
import trend_etl.schema.args as arguments


class TestBronzeTranforms(unittest.TestCase):
    spark: SparkSession

    @classmethod
    def setUpClass(cls) -> None:
        cls.spark = getspark()

    def test_kafka_get_groups(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "groups.jsonl")
        got_df = from_kafka_get_bronze_subreddits_df(source_df, "groups")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.collected_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_subreddits(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "subreddits.jsonl")
        got_df = from_kafka_get_bronze_groups_df(source_df, "subreddits")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.collected_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_posts(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "posts.jsonl")
        got_df = from_kafka_get_bronze_posts_df(source_df, "posts", reverse)

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_comments(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "comments.jsonl")
        got_df = from_kafka_get_bronze_comments_df(source_df, "comments", reverse)

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_submissions(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "submissions.jsonl")
        got_df = from_kafka_get_bronze_submissions_df(source_df, "submissions")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_replies(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "submissions.jsonl")
        got_df = from_kafka_get_bronze_replies_df(source_df, "replies")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_archive_get_groups(self) -> None:
        archive_df = get_archive_df(self.__class__.spark, "test_data", "groups.jsonl")
        group_test_id = "100"  # "200" is not a test statment

        got_df = from_archive_get_bronze_groups_df(archive_df, [group_test_id], False)
        self.assertEqual(1, got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.collected_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_df = from_archive_get_bronze_groups_df(archive_df, [group_test_id], True)
        self.assertEqual(1, got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.collected_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

    def test_archive_get_subreddits(self) -> None:
        archive_df = get_archive_df(
            self.__class__.spark, "test_data", "subreddits.jsonl"
        )

        got_df = from_archive_get_bronze_subreddits_df(archive_df)
        self.assertEqual(archive_df.count(), got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.collected_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

    def test_archive_get_posts(self) -> None:
        archive_df = get_archive_df(self.__class__.spark, "test_data", "posts.jsonl")
        group_test_id = "100"  # "200" is not a test statment

        got_df = from_archive_get_bronze_posts_df(
            archive_df, [group_test_id], False, reverse
        )
        self.assertEqual(1, got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_df = from_archive_get_bronze_posts_df(
            archive_df, [group_test_id], True, reverse
        )
        self.assertEqual(
            archive_df.count() - 1, got_df.count(), "dataframe size different"
        )

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

    def test_archive_get_comments(self) -> None:
        archive_df = get_archive_df(self.__class__.spark, "test_data", "comments.jsonl")
        group_test_id = "100"  # "200" is not a test statment

        got_df = from_archive_get_bronze_comments_df(
            archive_df, [group_test_id], False, reverse
        )
        self.assertEqual(4, got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_df = from_archive_get_bronze_comments_df(
            archive_df, [group_test_id], True, reverse
        )
        self.assertEqual(
            archive_df.count() - 4, got_df.count(), "dataframe size different"
        )

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

    def test_archive_get_submissions_reddit(self) -> None:
        archive_df = get_archive_df(
            self.__class__.spark, "test_data", "submissions.jsonl"
        )

        got_df = from_reddit_archive_get_bronze_submissions_df(archive_df)
        self.assertEqual(archive_df.count(), got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

    def test_archive_get_submissions_pushshift(self) -> None:
        archive_df = get_archive_df(
            self.__class__.spark, "test_data", "submissions.pushshift.jsonl"
        )

        got_df = from_pushshift_archive_get_bronze_submissions_df(archive_df)
        self.assertEqual(archive_df.count(), got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

    def test_archive_get_replies_reddit(self) -> None:
        archive_df = get_archive_df(self.__class__.spark, "test_data", "replies.jsonl")

        got_df = from_reddit_archive_get_bronze_replies_df(archive_df)
        self.assertEqual(archive_df.count(), got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

    def test_archive_get_replies_pushshift(self) -> None:
        archive_df = get_archive_df(
            self.__class__.spark, "test_data", "replies.pushshift.jsonl"
        )

        got_df = from_pushshift_archive_get_bronze_replies_df(archive_df)
        self.assertEqual(archive_df.count(), got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

    def test_facebook_get_idents_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "groups.jsonl")
        bronze_table = from_kafka_get_bronze_groups_df(source_df, "groups")
        got_df = from_bronze_get_idents_df(
            bronze_table, "$.id", arguments.tables[arguments.tables_group_idents]
        )

        got_columns = got_df.columns
        want_columns = [
            fields.idents_group_id_field,
            fields.idents_source_id_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_reddit_get_idents_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "subreddits.jsonl")
        bronze_table = from_kafka_get_bronze_subreddits_df(source_df, "unittest")
        got_df = from_bronze_get_idents_df(
            bronze_table,
            "$.data.name",
            arguments.tables[arguments.tables_subreddit_idents],
        )

        got_columns = got_df.columns
        want_columns = [
            fields.idents_group_id_field,
            fields.idents_source_id_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count() - 1  # skips kind = Listing
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_guilds(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "guilds.jsonl")
        got_df = from_kafka_get_bronze_guilds_df(source_df, "guilds")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_channels(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "channels.jsonl")
        got_df = from_kafka_get_bronze_channels_df(source_df, "channels")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_messages(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "messages.jsonl")
        got_df = from_kafka_get_bronze_messages_df(source_df, "messages")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_reactions(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "reactions.jsonl")
        got_df = from_kafka_get_bronze_reactions_df(source_df, "reactions")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.collected_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_edited_df(self) -> None:
        source_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "submissions.jsonl"
        )
        bronze_df = from_kafka_get_bronze_submissions_df(source_df, "submissions")

        parsed_df = bronze_df.select(
            parse_reddit_submissions(
                col(fields.bronze_value_field).cast(StringType())
            ).alias("value"),
            col(fields.bronze_timestamp_field),
        )
        flattened_df = parsed_df.select(
            col("value.data.*"),
            col(fields.bronze_timestamp_field),
        )

        got_df = get_edited_flag_dataframe(flattened_df, "name", "selftext")

        got_length = got_df.count()
        want_length = flattened_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

        edited_count = (
            got_df.drop_duplicates(["name", "subreddit_id", "_edited_flag"])
            .where(col("_edited_flag"))
            .count()
        )
        total_count = got_df.drop_duplicates(
            ["name", "subreddit_id", "_edited_flag"]
        ).count()

        self.assertEqual(edited_count, 1, "edited count size different")
        self.assertEqual(total_count, 5, "total count size different")

        total_count_by_subreddit = got_df.groupBy("subreddit_id").agg(
            size(collect_set("name")).alias("total_count")
        )

        edited_count_by_subreddit = (
            got_df.where(col("_edited_flag"))
            .groupBy("subreddit_id")
            .agg(size(collect_set("name")).alias("edited_count"))
        )

        # get the subreddits dataframe
        subreddits_df = from_bronze_get_subreddits_df(
            from_kafka_get_bronze_subreddits_df(
                get_kafka_df(self.__class__.spark, "test_data", "subreddits.jsonl"),
                "subreddits",
            ),
            get_df(
                self.__class__.spark,
                "group_id string, id string",
                "test_data",
                "subreddit_idents.jsonl",
            ),
        )

        got_df = (
            subreddits_df.selectExpr("source_keys.name as subreddit_id", "display_name")
            .join(total_count_by_subreddit, "subreddit_id", "left_outer")
            .join(edited_count_by_subreddit, "subreddit_id", "left_outer")
            .selectExpr(
                "subreddit_id",
                "coalesce(total_count, 0) as total_count",
                "coalesce(edited_count, 0) as edited_count",
            )
        )
        want_df = self.__class__.spark.createDataFrame(
            [
                ("t5_qn4wb", 2, 1),
                ("t5_zn4wb", 1, 0),
                ("t5_zn4wr", 1, 0),
                ("t5_zn4qq", 1, 0),
            ],
            schema="subreddit_id string, total_count int not null, edited_count int not null",
        )

        assert_dataframes_equal(got_df, want_df)
