from pyspark.sql import DataFrame, Column

import pyspark.sql.functions as fn

import trend_etl.schema.args as arguments
from trend_etl.functions.udfs import conv_timestamp

##
## conditions
##


def is_trend_employee(user_id_field: str) -> Column:
    return fn.col(user_id_field).isin(arguments.discord_trend_employees)


def is_bot_content_flag() -> Column:
    return fn.col("bot_content_flag")


def is_bot() -> Column:
    return fn.coalesce(fn.col("bot"), fn.lit(False))


def days_ago_cond(timestamp: Column, days_ago: int) -> Column:
    return fn.coalesce(
        fn.date_add(fn.current_timestamp(), -days_ago) <= timestamp, fn.lit(False)
    )


def greater_days_cond(timestamp: Column, days_ago: int) -> Column:
    return fn.coalesce(
        fn.date_add(fn.current_timestamp(), -days_ago) > timestamp, fn.lit(False)
    )


##
## stat functions
##


# total registered users: get all members for all guilds
def total_registered_users_number(users_table: DataFrame) -> int:
    unique_users = users_table.drop_duplicates(["source_keys"])
    return unique_users.where(~is_trend_employee("source_keys.id") & ~is_bot()).count()


# total journeys: total # of journeys, specifically posts to the corresponding
#                 jounreys channel (posts are threads)
# NOTE: calling cache() was need to get tests to pass.
#       when upgrade to subsequent version see if we can remove calling cache()
#       with the test passing (OOM error)
def total_journeys_number(
    channels_table: DataFrame, users_table: DataFrame, parent_channel_ids: list[str]
) -> int:
    # fetch threads (channels) that have the journeys channel as its parent
    journeys = channels_table.where(
        fn.col("source_keys.parent_id").isin(parent_channel_ids)
    )

    bots = users_table.where(is_bot())

    filtered = (
        journeys
        # filter out journeys owned by a bot user (probably would never occur)
        .join(
            bots.selectExpr("source_keys.id as user_id"),
            fn.col("user_id") == fn.col("source_keys.thread_owner_id"),
            "left_anti",
        ).cache()
    )

    return filtered.where(~is_trend_employee("source_keys.thread_owner_id")).count()


# journeys conversion: number of registered users that created a journey
def journey_conversion_number(
    channels_table: DataFrame,
    users_table: DataFrame,
    parent_channel_ids: list[str],
) -> int:
    # fetch threads (channels) that have the journeys channel as its parent
    journeys = channels_table.where(
        fn.col("source_keys.parent_id").isin(parent_channel_ids)
    )

    unique_users = users_table.drop_duplicates(["source_keys"])
    filtered_registered_users = unique_users.where(
        ~is_trend_employee("source_keys.id") & ~is_bot()
    )

    # get registered users that have created a journey
    conversions = filtered_registered_users.alias("user").join(
        journeys.alias("journeys"),
        fn.col("journeys.source_keys.thread_owner_id") == fn.col("user.source_keys.id"),
        "left_semi",
    )

    return conversions.count()


# total statments: total # statments in all for all discord guilds (threads + messages)
def total_statements_number(threads_table: DataFrame, messages_table: DataFrame) -> int:
    # get count of all threads that are not trend employees
    threads_filtered = threads_table.where(
        ~is_trend_employee("source_keys.thread_owner_id")
    )
    threads_filtered_number = threads_filtered.count()

    # get a count all messages that are not trend employees
    messages = messages_table.where(
        ~is_trend_employee("source_keys.user_id") & ~is_bot_content_flag()
    )
    messages_number = messages.count()

    return threads_filtered_number + messages_number


# active users: # of users who have posted or replied in at least `message_count_threshold`
#               times in the last 30, 60, 90 days
def active_users_days_ago_number(
    threads_table: DataFrame,
    messages_table: DataFrame,
    days_ago: int,
    message_count_threshold: int,
) -> int:
    thread_created = conv_timestamp("thread_metadata.create_timestamp")
    message_created = fn.col("created_at")

    active_users = active_users_for_threads(
        threads_table, days_ago_cond(thread_created, days_ago)
    ).unionAll(
        active_users_for_messages(
            messages_table, days_ago_cond(message_created, days_ago)
        )
    )

    active_users_num = (
        active_users.groupBy("user_id")
        .count()
        .where(fn.col("count") >= message_count_threshold)
        .count()
    )

    return active_users_num


# active audience: # of users who have posted or replied in at least `message_count_threshold`
#                  times in the last `days_ago` days PLUS # of users who have reacted at least
#                  `reaction_count_threshold` in the last `days_ago` days.
#                  Essentially Active Users + "lurkers"
def active_audience_days_ago_number(
    threads_table: DataFrame,
    messages_table: DataFrame,
    reactions_table: DataFrame,
    users_table: DataFrame,
    days_ago: int,
    message_count_threshold: int,
    reaction_count_threshold: int,
) -> int:
    thread_created = conv_timestamp("thread_metadata.create_timestamp")
    message_created = fn.col("created_at")

    # Get all active users using the `days_ago` condition
    active_users = active_users_for_threads(
        threads_table, days_ago_cond(thread_created, days_ago)
    ).unionAll(
        active_users_for_messages(
            messages_table, days_ago_cond(message_created, days_ago)
        )
    )

    # Get users that reacted using the `days_ago` condition and,
    # filter out users already accounted for in `active_users`. In other words,
    # this collects all users that reacted that are not in the `active_users`
    # collection.
    active_users_reactions = active_users_for_reactions(
        messages_table,
        reactions_table,
        users_table,
        days_ago_cond(message_created, days_ago),
    ).join(active_users, "user_id", "left_anti")

    # active users + reactions = active audience
    # (uses the appropriate thresholds)

    active_users_num = (
        active_users.groupBy("user_id")
        .count()
        .where(fn.col("count") >= message_count_threshold)
        .count()
    )

    active_users_reactions_num = (
        active_users_reactions.groupBy("user_id")
        .count()
        .where(fn.col("count") >= reaction_count_threshold)
        .count()
    )

    return active_users_num + active_users_reactions_num


# Retention:
#   The number of users who interacted within <=`days_partition` day ago
#   on the server, who also have interacted on the server
#   >`days_partition` days ago
def retention_audience_number(
    threads_table: DataFrame,
    messages_table: DataFrame,
    reactions_table: DataFrame,
    users_table: DataFrame,
    days_partition: int,
) -> int:
    thread_created = conv_timestamp("thread_metadata.create_timestamp")
    message_created = fn.col("created_at")

    # get users (dupes likely) that make up the active audience for
    # <=`days_partition` days ago
    active_audience_users_recent = (
        active_users_for_threads(
            threads_table, days_ago_cond(thread_created, days_partition)
        )
        .unionAll(
            active_users_for_messages(
                messages_table, days_ago_cond(message_created, days_partition)
            )
        )
        .unionAll(
            active_users_for_reactions(
                messages_table,
                reactions_table,
                users_table,
                days_ago_cond(message_created, days_partition),
            )
        )
    ).cache()

    # get users (dupes likely) that make up the active audience for
    # >`days_partition` days ago
    active_audience_users_former = (
        active_users_for_threads(
            threads_table, greater_days_cond(thread_created, days_partition)
        )
        .unionAll(
            active_users_for_messages(
                messages_table, greater_days_cond(message_created, days_partition)
            )
        )
        .unionAll(
            active_users_for_reactions(
                messages_table,
                reactions_table,
                users_table,
                greater_days_cond(message_created, days_partition),
            )
        )
    ).cache()

    audience_retention_users = active_audience_users_recent.intersect(
        active_audience_users_former
    )

    return audience_retention_users.count()


##
## stats helpers
##


# returns: DataFrame with single column `user_id`
def active_users_for_threads(
    threads_table: DataFrame, threads_date_cond: Column
) -> DataFrame:
    # get count of all threads that are not trend employees
    threads_filtered = threads_table.where(
        ~is_trend_employee("source_keys.thread_owner_id")
    )

    # get all threads that meet the date condition
    threads_days = threads_filtered.where(threads_date_cond)

    return threads_days.selectExpr("source_keys.thread_owner_id as user_id")


# returns: DataFrame with single column `user_id`
def active_users_for_messages(
    messages_table: DataFrame, messages_date_cond: Column
) -> DataFrame:
    messages_filtered = messages_table.where(
        ~is_trend_employee("source_keys.user_id") & ~is_bot_content_flag()
    )

    # get all messages that meet the date condition
    messages_days = messages_filtered.where(messages_date_cond)

    return messages_days.selectExpr("source_keys.user_id as user_id")


# returns: DataFrame with single column `user_id`
# NOTE: calling cache() was need to get tests to pass.
#       when upgrade to subsequent version see if we can remove calling cache()
#       with the test passing (OOM error)
def active_users_for_reactions(
    messages_table: DataFrame,
    reactions_table: DataFrame,
    users_table: DataFrame,
    messages_date_cond: Column,
) -> DataFrame:
    # NOTE: we don't want to filter trend employee messages since a
    #       non- trend employee may react and that should count too.

    # get all messages that meet the date condition
    messages_days = messages_table.where(messages_date_cond)

    # get all the reactions for all messages
    messages_reactions = reactions_table.alias("reactions").join(
        messages_days.alias("messages"),
        fn.col("messages.source_keys.id") == fn.col("reactions.source_keys.message_id"),
        "left_semi",
    )

    # filter out reactions from trend employees or bots
    bots = users_table.where(is_bot())

    filtered = (
        messages_reactions
        # filter out bot reactions (probably would never occur)
        .join(
            bots.selectExpr("source_keys.id as user_id"),
            fn.col("user_id") == fn.col("source_keys.user_id"),
            "left_anti",
        ).cache()
    )

    reactions_filtered = filtered.where(~is_trend_employee("source_keys.user_id"))

    return reactions_filtered.selectExpr("source_keys.user_id as user_id")
