from pyspark.sql import DataFrame
from typing import List

# from pyspark.sql.column import Column
import pyspark.sql.functions as fn
from pyspark.sql.types import StringType, BinaryType

import trend_etl.schema.fields as fields
from trend_etl.functions.misc import (
    jobid_2_timestamp,
    get_collected_day,
)
from trend_etl.functions.udfs import conv_timestamp
from trend_etl.functions.parsing import parse_facebook_groups
import trend_etl.functions.ingest_versioning as ingest
from trend_etl.schema.canonical import facebook_group_canonical


# arguments:
#   archive_df: the dataframe from the archives
#   group_ids: the group ids used in filtering
#   include_flag: whether to use the filtering logic to include/exclude
#   provenance: short description of the origin (kafka, s3 archive, etc.)
def from_archive_get_bronze_groups_df(
    archive_df: DataFrame, group_ids: List[str], include_flag: bool
) -> DataFrame:
    field = fn.col(fields.archive_value_field)

    return (
        archive_df.select(
            fn.get_json_object(field, "$.id").isin(group_ids).alias("_filter"),
            fn.col(fields.archive_value_field)
            .cast(BinaryType())
            .alias(fields.bronze_value_field),
            fn.lit("archive/facebook/groups").alias(fields.bronze_provenance_field),
            jobid_2_timestamp(fields.archive_job_id_field).alias(
                fields.bronze_timestamp_field
            ),
        )
        .where(fn.col("_filter") == include_flag)
        .drop("_filter")
        .withColumn(
            fields.collected_day_field,
            get_collected_day(fn.col(fields.bronze_timestamp_field)),
        )
    )


# arguments:
#   input_df: dataframe received from reading a kafka topic
#   target_topic: source topic from kafka
# return: a dataframe in the schema expected for bronze tables
def from_kafka_get_bronze_groups_df(
    input_df: DataFrame, target_topic: str
) -> DataFrame:
    return input_df.select(
        fn.col(fields.kafka_value_field).alias(fields.bronze_value_field),
        fn.lit(f"kafka/{target_topic}").alias(fields.bronze_provenance_field),
        fn.col(fields.kafka_timestamp_field).alias(fields.bronze_timestamp_field),
        get_collected_day(fn.col(fields.kafka_timestamp_field)).alias(
            fields.collected_day_field
        ),
    )


# arguments:
#   streaming_bronze: streaming dataframe from bronze table
#   group_idents: groups uuid idents table
# return: dataframe in the schema of facebook_groups
def from_bronze_get_groups_df(
    streaming_bronze: DataFrame,
    group_idents: DataFrame,
) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_facebook_groups(
            fn.col(fields.bronze_value_field).cast(StringType())
        ).alias("value"),
        fn.col(fields.bronze_timestamp_field),
        fn.col(fields.collected_day_field),
    )
    flattened_df = parsed_df.select(
        fn.col("value.*"),
        fn.col(fields.bronze_timestamp_field),
        fn.col(fields.collected_day_field),
    )

    # NOTE: left outer join (below) will cause any nulls to be
    #       detected as a downstream error.

    # join with group_idents
    group_idents_df = flattened_df.join(
        group_idents,
        fields.idents_source_id_field,
        "left_outer",
    )

    groups_df = group_idents_df.select(
        fn.col(fields.idents_group_id_field),
        fn.col("email").alias(fields.facebook_groups_email),
        fn.col("description").alias(fields.facebook_groups_description),
        conv_timestamp("created_time").alias(fields.facebook_groups_created_at),
        fn.col("name").alias(fields.facebook_groups_name),
        fn.col("privacy").alias(fields.facebook_groups_privacy),
        fn.col("member_count").alias(fields.facebook_groups_member_count),
        fn.struct(fn.col("id").alias("id")).alias(fields.facebook_groups_source_keys),
        fn.col(fields.bronze_timestamp_field),
        fn.col(fields.collected_day_field),
    )

    return groups_df


def from_bronze_get_groups_df_ao(
    streaming_bronze: DataFrame,
    group_idents: DataFrame,
) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_facebook_groups(
            fn.col(fields.bronze_value_field).cast(StringType())
        ).alias("json"),
        fn.col(fields.bronze_timestamp_field).alias(fields.ingested_on_field),
    )

    flattened_df = parsed_df.select(
        ingest.ingest_version(fn.col("json"), facebook_group_canonical).alias(
            fields.ingest_version_field
        ),
        fn.col("json.*"),
        fn.col(fields.ingested_on_field),
    )

    # NOTE: left outer join (below) will cause any nulls to be
    #       detected as a downstream error.

    # join with group_idents
    group_idents_df = flattened_df.join(
        group_idents,
        fields.idents_source_id_field,
        "left_outer",
    )

    result_df = group_idents_df.select(
        fn.col(fields.ingest_version_field),
        fn.col(fields.idents_group_id_field),
        fn.struct(fn.col("id")).alias(fields.source_keys_field),
        fn.col(fields.ingested_on_field),
        fn.col("email").alias(fields.facebook_groups_email),
        fn.col("description").alias(fields.facebook_groups_description),
        fn.col("name").alias(fields.facebook_groups_name),
        fn.col("privacy").alias(fields.facebook_groups_privacy),
        fn.col("member_count").alias(fields.facebook_groups_member_count),
        conv_timestamp("created_time").alias(fields.facebook_groups_created_at),
    )

    return result_df
