from pyspark.sql import DataFrame

import pyspark.sql.functions as fn
from pyspark.sql.types import (
    StringType,
    BinaryType,
    TimestampType,
    LongType,
    DoubleType,
)

import trend_etl.schema.fields as fields
import trend_etl.schema.args as arguments
from trend_etl.functions.misc import (
    jobid_2_timestamp,
    get_created_day,
    is_empty_or_white_space_reddit,
)
from trend_etl.functions.parsing import parse_reddit_replies, parse_pushshift_replies
from trend_etl.functions.udfs import is_scientific
import trend_etl.functions.ingest_versioning as ingest
from trend_etl.schema.canonical import (
    reddit_reply_canonical,
    reddit_reply_text_canonical,
)


def from_reddit_archive_get_bronze_replies_df(archive_df: DataFrame) -> DataFrame:
    parsed = archive_df.withColumn(
        fields.archive_value_field,
        parse_reddit_replies(fn.col(fields.archive_value_field)),
    )
    created_utc_field = fn.col(f"{fields.archive_value_field}.data.created_utc")

    return parsed.select(
        fn.to_json(fn.col(fields.archive_value_field))
        .cast(BinaryType())
        .alias(fields.bronze_value_field),
        fn.lit("archive/reddit/replies").alias(fields.bronze_provenance_field),
        jobid_2_timestamp(fields.archive_job_id_field).alias(
            fields.bronze_timestamp_field
        ),
        get_created_day(created_utc_field.cast(TimestampType())).alias(
            fields.created_day_field
        ),
    )


def from_pushshift_archive_get_bronze_replies_df(
    archive_df: DataFrame,
) -> DataFrame:
    parsed = archive_df.withColumn(
        fields.archive_value_field,
        parse_pushshift_replies(fn.col(fields.archive_value_field)),
    )
    id_field = fn.col(f"{fields.archive_value_field}.id")
    created_utc_field = fn.col(f"{fields.archive_value_field}.created_utc")

    return (
        parsed.withColumn(
            fields.archive_value_field,
            fn.col(fields.archive_value_field).withField(
                "name",
                fn.format_string("t1_%s", id_field),
            ),
        )
        .withColumn(
            fields.archive_value_field,
            fn.col(fields.archive_value_field).withField(
                "created_utc",
                created_utc_field.cast(LongType()),
            ),
        )
        .select(
            fn.to_json(
                fn.struct(
                    fn.lit("t1").alias("kind"),
                    fn.col(fields.archive_value_field).alias("data"),
                    fn.lit("pushshift_archive").alias("note"),
                )
            )
            .cast(BinaryType())
            .alias(fields.bronze_value_field),
            fn.lit("archive/pushshift/replies").alias(fields.bronze_provenance_field),
            jobid_2_timestamp(fields.archive_job_id_field).alias(
                fields.bronze_timestamp_field
            ),
            get_created_day(created_utc_field.cast(TimestampType())).alias(
                fields.created_day_field
            ),
        )
    )


# arguments:
#   input_df: dataframe received from reading a kafka topic
#   target_topic: source topic from kafka
# return: a dataframe in the schema expected for bronze tables (for all reddit data kinds)
def from_kafka_get_bronze_replies_df(input_df: DataFrame, target_topic: str):
    parsed = input_df.withColumn(
        fields.kafka_value_field,
        parse_reddit_replies(fn.col(fields.kafka_value_field).cast(StringType())),
    )
    created_utc_field = fn.col(f"{fields.archive_value_field}.data.created_utc")

    return parsed.select(
        fn.to_json(fn.col(fields.kafka_value_field))
        .cast(BinaryType())
        .alias(fields.bronze_value_field),
        fn.lit(f"kafka/{target_topic}").alias(fields.bronze_provenance_field),
        fn.col(fields.kafka_timestamp_field).alias(fields.bronze_timestamp_field),
        get_created_day(created_utc_field.cast(TimestampType())).alias(
            fields.created_day_field
        ),
    )


# arguments:
#   streaming_bronze: streaming dataframe from bronze table
#   subreddit_idents: subreddit uuid idents table
#   submission_idents: submission uuid idents table
#   reply_idents: reply uuid idents table
#   bot_authors: reddit bot users table
# return: dataframe in the schema of `reddit_replies`
def from_bronze_get_replies_df(
    streaming_bronze: DataFrame,
    subreddit_idents: DataFrame,
    submission_idents: DataFrame,
    reply_idents: DataFrame,
    bot_authors: DataFrame,
) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_reddit_replies(
            fn.col(fields.bronze_value_field).cast(StringType())
        ).alias("value"),
        fn.col(fields.bronze_timestamp_field),
    )
    flattened_df = parsed_df.select(
        fn.col("value.data.*"),
        fn.col(fields.bronze_timestamp_field),
    )

    # filter out bot submissions
    bot_filtered_df = flattened_df.join(
        bot_authors, fields.reddit_bot_users_author, "left_anti"
    )

    test_filtered_df = bot_filtered_df.where(
        fn.col("subreddit_id").isin(arguments.test_subreddits) == False
    )

    # NOTE: left outer joins (below) will cause any nulls to be
    #       detected as a downstream error.

    # join with subreddit_idents
    subreddits_df = test_filtered_df.join(
        subreddit_idents.withColumnRenamed(
            fields.idents_source_id_field, "subreddit_id"
        ),
        "subreddit_id",
        "inner",  # filter out replies from subreddit_type = 'user' and like which we don't track
    ).drop("subreddit_id")

    # join with reply_idents to get doc_uuid
    reply_idents_df = subreddits_df.join(
        reply_idents.withColumnRenamed(fields.idents_source_id_field, "name"),
        "name",
        "left_outer",
    ).withColumnRenamed("name", "_source_key_id")

    # join with submission_idents to get corresponding submission uuid fk
    assoc_submission_df = reply_idents_df.join(
        submission_idents.withColumnRenamed(
            fields.idents_doc_id_field, fields.reddit_replies_submission_uuid
        ).withColumnRenamed(fields.idents_source_id_field, "link_id"),
        "link_id",
        "left_outer",
    ).withColumnRenamed("link_id", "_source_key_link_id")

    # join with comment_idents to get parent uuid fk
    assoc_parent_df = assoc_submission_df.withColumnRenamed(
        "parent_id", "_source_key_parent_id"
    ).join(
        reply_idents.withColumnRenamed(
            fields.idents_doc_id_field, fields.reddit_replies_parent_uuid
        ).withColumnRenamed(fields.idents_source_id_field, "_source_key_parent_id"),
        "_source_key_parent_id",
        "left_outer",
    )

    return (
        assoc_parent_df.withColumn(
            fields.reddit_replies_created_at,
            fn.when(
                is_scientific(fn.col("created_utc")),
                fn.col("created_utc").cast(DoubleType()).cast(LongType()),
            )
            .otherwise(fn.col("created_utc").cast(LongType()))
            .cast(TimestampType()),
        )
        .drop("created_utc")
        .withColumnRenamed("body", fields.reddit_replies_text)
        .withColumn(
            fields.source_keys_field,
            fn.struct(
                fn.col("_source_key_id").alias("name"),
                fn.col("_source_key_link_id").alias("submission_id"),
                fn.col("_source_key_parent_id").alias("parent_id"),
            ),
        )
        .drop("_source_key_id", "_source_key_link_id", "_source_key_parent_id")
    )


def from_bronze_get_replies_df_append(
    streaming_bronze: DataFrame,
    subreddit_idents: DataFrame,
    submission_idents: DataFrame,
    reply_idents: DataFrame,
) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_reddit_replies(
            fn.col(fields.bronze_value_field).cast(StringType())
        ).alias("value"),
        fn.col(fields.bronze_timestamp_field),
    )
    flattened_df = parsed_df.select(
        fn.col("value.data.*"),
        fn.col(fields.bronze_timestamp_field),
        fn.when(
            fn.coalesce(fn.col("value.note"), fn.lit("reddit")) == "pushshift_archive",
            "pushshift",
        )
        .otherwise("reddit")
        .alias("source_schema"),
    )

    # NOTE: `inner` join (below) will filter out submissions from
    # subreddit_type = 'user' and like which we don't track

    # join with subreddit_idents
    with_subreddit_idents = flattened_df.join(
        subreddit_idents.withColumnRenamed(
            fields.idents_source_id_field, "subreddit_id"
        ),
        "subreddit_id",
        "inner",
    ).drop("subreddit_id")

    # NOTE: `left_outer` join (below) will cause any nulls to be
    #       detected as a downstream error.

    with_reply_idents = with_subreddit_idents.join(
        reply_idents.withColumnRenamed(fields.idents_source_id_field, "name"),
        "name",
        "left_outer",
    ).withColumnRenamed("name", "_source_key_id")

    with_submission_idents = with_reply_idents.join(
        submission_idents.withColumnRenamed(
            fields.idents_doc_id_field, fields.reddit_replies_submission_uuid
        ).withColumnRenamed(fields.idents_source_id_field, "link_id"),
        "link_id",
        "left_outer",
    ).withColumnRenamed("link_id", "_source_key_link_id")

    with_parent_set = with_submission_idents.withColumnRenamed(
        "parent_id", "_source_key_parent_id"
    ).join(
        reply_idents.withColumnRenamed(
            fields.idents_doc_id_field, fields.reddit_replies_parent_uuid
        ).withColumnRenamed(fields.idents_source_id_field, "_source_key_parent_id"),
        "_source_key_parent_id",
        "left_outer",
    )

    return (
        with_parent_set.withColumn(
            fields.reddit_replies_created_at,
            fn.when(
                is_scientific(fn.col("created_utc")),
                fn.col("created_utc").cast(DoubleType()).cast(LongType()),
            )
            .otherwise(fn.col("created_utc").cast(LongType()))
            .cast(TimestampType()),
        )
        .drop("created_utc")
        .withColumnRenamed("body", fields.reddit_replies_text)
        .withColumn(
            fields.source_keys_field,
            fn.struct(
                fn.col("_source_key_id").alias("name"),
                fn.col("_source_key_link_id").alias("submission_id"),
                fn.col("_source_key_parent_id").alias("parent_id"),
            ),
        )
        .drop("_source_key_id", "_source_key_link_id", "_source_key_parent_id")
    )


def from_bronze_get_replies_df_ao(
    streaming_bronze: DataFrame,
    reply_idents: DataFrame,
) -> DataFrame:
    parsed = streaming_bronze.select(
        parse_reddit_replies(
            fn.col(fields.bronze_value_field).cast(StringType())
        ).alias("json"),
        fn.col(fields.bronze_timestamp_field).alias(fields.ingested_on_field),
    )

    flattened = parsed.select(
        ingest.ingest_version(
            fn.col("json").getField("data"), reddit_reply_canonical
        ).alias(fields.ingest_version_field),
        fn.when(
            ~is_empty_or_white_space_reddit(
                fn.coalesce(
                    fn.col("json").getField("data").getField("body"), fn.lit("")
                )
            ),
            ingest.ingest_version(
                fn.col("json").getField("data"), reddit_reply_text_canonical
            ),
        ).alias(fields.text_version_field),
        fn.col("json.data.*"),
        fn.col(fields.ingested_on_field),
        fn.when(
            fn.coalesce(fn.col("json.note"), fn.lit("reddit")) == "pushshift_archive",
            "pushshift",
        )
        .otherwise("reddit")
        .alias("source_schema"),
    )

    return (
        flattened.join(
            reply_idents.withColumnRenamed(fields.idents_source_id_field, "name"),
            "name",
            "left_outer",
        )
        .withColumn(
            fields.source_keys_field,
            fn.struct(
                fn.col("name"),
                fn.col("link_id").alias("submission_id"),
                # NOTE: Will use this info to differentiate between missing and null cases
                fn.col("parent_id"),
                # fn.when(
                #     fn.col("parent_id") != fn.col("link_id"), fn.col("parent_id")
                # ).alias("parent_id"),
                fn.col("author"),
                fn.col("author_fullname"),
                fn.col("subreddit_id"),
            ),
        )
        .withColumn(
            fields.reddit_replies_created_at,
            fn.when(
                is_scientific(fn.col("created_utc")),
                fn.col("created_utc").cast(DoubleType()).cast(LongType()),
            )
            .otherwise(fn.col("created_utc").cast(LongType()))
            .cast(TimestampType()),
        )
        .drop(
            "created_utc",
            "name",
            "link_id",
            "parent_id",
            "author",
            "author_fullname",
            "subreddit_id",
            "body",
        )
    )


def from_bronze_get_replies_text_df_ao(streaming_bronze: DataFrame) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_reddit_replies(
            fn.col(fields.bronze_value_field).cast(StringType())
        ).alias("json")
    )

    result_df = parsed_df.select(
        ingest.ingest_version(
            fn.col("json").getField("data"), reddit_reply_text_canonical
        ).alias(fields.text_version_field),
        fn.col("json").getField("data").getField("body").alias("text"),
    ).where(
        ~is_empty_or_white_space_reddit(
            fn.coalesce(fn.col("json").getField("data").getField("body"), fn.lit(""))
        )
    )

    return result_df
