import unittest

from pyspark.sql import SparkSession
import pyspark.sql.functions as fn

from trend_etl.test_utils import getspark, get_kafka_df, get_df
from trend_etl.transform.groups import (
    from_kafka_get_bronze_groups_df,
    from_bronze_get_groups_df_ao,
)
from trend_etl.transform.posts import (
    from_kafka_get_bronze_posts_df,
    from_bronze_get_posts_df_ao,
    from_bronze_get_posts_text_df_ao,
)
from trend_etl.transform.comments import (
    from_kafka_get_bronze_comments_df,
    from_bronze_get_comments_df_ao,
    from_bronze_get_comments_text_df_ao,
)
from trend_etl.transform.subreddits import (
    from_kafka_get_bronze_subreddits_df,
    from_bronze_get_subreddits_df_ao,
)
from trend_etl.transform.submissions import (
    from_kafka_get_bronze_submissions_df,
    from_bronze_get_submissions_df,
    from_bronze_get_submissions_df_ao,
    from_bronze_get_submissions_text_df_ao,
)
from trend_etl.transform.replies import (
    from_kafka_get_bronze_replies_df,
    from_bronze_get_replies_df,
    from_bronze_get_replies_df_ao,
    from_bronze_get_replies_text_df_ao,
)

from trend_etl.transform.community_groups import from_bronze_get_community_groups_df

from trend_etl.transform.guilds import (
    from_bronze_get_guilds_df_ao,
    from_kafka_get_bronze_guilds_df,
)
from trend_etl.transform.channels import (
    from_bronze_get_channels_df_ao,
    from_kafka_get_bronze_channels_df,
)
from trend_etl.transform.messages import (
    from_bronze_get_messages_df_ao,
    from_bronze_get_discord_text_df_ao,
    from_bronze_get_discord_mention_user_df_ao,
    from_bronze_get_discord_mention_channel_df_ao,
    from_kafka_get_bronze_messages_df,
)
from trend_etl.transform.reactions import (
    from_bronze_get_reactions_df_ao,
    from_kafka_get_bronze_reactions_df,
)
from trend_etl.transform.members import (
    from_bronze_get_members_df_ao,
    from_bronze_get_users_df_ao,
    from_kafka_get_bronze_members_df,
)

import trend_etl.schema.fields as fields


class TestSilverTranforms(unittest.TestCase):
    spark: SparkSession

    @classmethod
    def setUpClass(cls) -> None:
        cls.spark = getspark()

    def test_get_groups_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "groups.jsonl")
        group_idents = get_df(
            self.__class__.spark,
            "group_id string, id string",
            "test_data",
            "group_idents.jsonl",
        )
        got_df = from_bronze_get_groups_df_ao(
            from_kafka_get_bronze_groups_df(source_df, "unittest"), group_idents
        )

        got_columns = got_df.columns
        want_columns = [
            "ingest_version",
            "group_id",
            "source_keys",
            "ingested_on",
            "email",
            "description",
            "name",
            "privacy",
            "member_count",
            "created_at",
        ]
        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_posts_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "posts.jsonl")
        post_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "post_idents.jsonl",
        )
        got_df = from_bronze_get_posts_df_ao(
            from_kafka_get_bronze_posts_df(source_df, "posts", fn.reverse),
            post_idents,
        )

        got_columns = got_df.columns
        want_columns = [
            "ingest_version",
            "text_version",
            "doc_uuid",
            "source_keys",
            "ingested_on",
            "permalink_url",
            "reactions",
            "type",
            "title",
            "link",
            "story",
            "created_at",
            "updated_at",
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_posts_text_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "posts.jsonl")
        got_df = from_bronze_get_posts_text_df_ao(
            from_kafka_get_bronze_posts_df(source_df, "posts", fn.reverse),
        )

        got_columns = got_df.columns
        want_columns = [
            "text_version",
            "text",
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count() - 1  # 1 with record with null message
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_comments_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "comments.jsonl")
        comment_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "comment_idents.jsonl",
        )
        got_df = from_bronze_get_comments_df_ao(
            from_kafka_get_bronze_comments_df(source_df, "comments", fn.reverse),
            comment_idents,
        )

        got_columns = got_df.columns
        want_columns = [
            "ingest_version",
            "text_version",
            "doc_uuid",
            "source_keys",
            "ingested_on",
            "permalink_url",
            "reactions",
            "created_at",
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_comments_text_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "comments.jsonl")
        got_df = from_bronze_get_comments_text_df_ao(
            from_kafka_get_bronze_comments_df(source_df, "comments", fn.reverse),
        )

        got_columns = got_df.columns
        want_columns = [
            "text_version",
            "text",
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()  # 0 records with null message
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_subreddits_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "subreddits.jsonl")
        subreddit_idents = get_df(
            self.__class__.spark,
            "group_id string, id string",
            "test_data",
            "subreddit_idents.jsonl",
        )
        got_df = from_bronze_get_subreddits_df_ao(
            from_kafka_get_bronze_subreddits_df(source_df, "subreddits"),
            subreddit_idents,
        )

        got_columns = got_df.columns
        want_columns = [  # copied from table bundle/src/setup_prod_catalog.ipynb
            "ingest_version",
            "group_id",
            "source_keys",
            "ingested_on",
            "accept_followers",
            "accounts_active",
            "accounts_active_is_fuzzed",
            "active_user_count",
            "advertiser_category",
            "all_original_content",
            "allow_chat_post_creation",
            "allow_discovery",
            "allow_galleries",
            "allow_images",
            "allow_polls",
            "allow_prediction_contributors",
            "allow_predictions",
            "allow_predictions_tournament",
            "allow_talks",
            "allow_videogifs",
            "allow_videos",
            "allowed_media_in_comments",
            "banner_background_color",
            "banner_background_image",
            "banner_img",
            "banner_size",
            "can_assign_link_flair",
            "can_assign_user_flair",
            "collapse_deleted_comments",
            "comment_contribution_settings",
            "comment_score_hide_mins",
            "community_icon",
            "community_reviewed",
            "description",
            "description_html",
            "disable_contributor_requests",
            "display_name",
            "display_name_prefixed",
            "emojis_custom_size",
            "emojis_enabled",
            "free_form_reports",
            "has_menu_widget",
            "header_img",
            "header_size",
            "header_title",
            "hide_ads",
            "icon_img",
            "icon_size",
            "interstitial_warning_message",
            "is_chat_post_feature_enabled",
            "is_crosspostable_subreddit",
            "is_enrolled_in_new_modmail",
            "key_color",
            "lang",
            "link_flair_enabled",
            "link_flair_position",
            "mobile_banner_image",
            "notification_level",
            "original_content_tag_enabled",
            "over18",
            "prediction_leaderboard_entry_type",
            "primary_color",
            "public_description",
            "public_description_html",
            "public_traffic",
            "quarantine",
            "restrict_commenting",
            "restrict_posting",
            "should_archive_posts",
            "should_show_media_in_comments_setting",
            "show_media",
            "show_media_preview",
            "spoilers_enabled",
            "submission_type",
            "submit_link_label",
            "submit_text",
            "submit_text_html",
            "submit_text_label",
            "subreddit_type",
            "subscribers",
            "suggested_comment_sort",
            "title",
            "url",
            "user_can_flair_in_sr",
            "user_flair_background_color",
            "user_flair_css_class",
            "user_flair_enabled_in_sr",
            "user_flair_position",
            "user_flair_richtext",
            "user_flair_template_id",
            "user_flair_text",
            "user_flair_text_color",
            "user_flair_type",
            "user_has_favorited",
            "user_is_banned",
            "user_is_contributor",
            "user_is_moderator",
            "user_is_muted",
            "user_is_subscriber",
            "user_sr_flair_enabled",
            "user_sr_theme_enabled",
            "videostream_links_count",
            "whitelist_status",
            "wiki_enabled",
            "wls",
            "created_at",
        ]

        self.assertEqual(set(want_columns), set(got_columns), "columns match error")

        got_length = got_df.count()
        want_length = source_df.count() - 1  # skips kind = Listing
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_submissions_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "submissions.jsonl")
        subreddit_idents = get_df(
            self.__class__.spark,
            "group_id string, id string",
            "test_data",
            "subreddit_idents.jsonl",
        )
        submission_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "submission_idents.jsonl",
        )
        bots = get_df(
            self.__class__.spark,
            "author string",
            "test_data",
            "reddit_bots.jsonl",
        )
        got_df = from_bronze_get_submissions_df(
            from_kafka_get_bronze_submissions_df(source_df, "submissions"),
            subreddit_idents,
            submission_idents,
            bots,
        )

        got_columns = got_df.columns
        want_columns = [
            "doc_uuid",
            "group_id",
            "created_at",
            "approved_at_utc",
            "text",
            "author_fullname",
            "saved",
            "mod_reason_title",
            "gilded",
            "clicked",
            "title",
            "link_flair_richtext",
            "hidden",
            "pwls",
            "link_flair_css_class",
            "downs",
            "thumbnail_height",
            "top_awarded_type",
            "hide_score",
            "quarantine",
            "link_flair_text_color",
            "upvote_ratio",
            "author_flair_background_color",
            "subreddit_type",
            "ups",
            "total_awards_received",
            "thumbnail_width",
            "author_flair_template_id",
            "is_original_content",
            "user_reports",
            "is_reddit_media_domain",
            "is_meta",
            "category",
            "link_flair_text",
            "can_mod_post",
            "score",
            "approved_by",
            "is_created_from_ads_ui",
            "author_premium",
            "thumbnail",
            "edited",
            "author_flair_css_class",
            "author_flair_richtext",
            "gildings",
            "post_hint",
            "content_categories",
            "is_self",
            "mod_note",
            "link_flair_type",
            "wls",
            "removed_by_category",
            "banned_by",
            "author_flair_type",
            "domain",
            "allow_live_comments",
            "likes",
            "suggested_sort",
            "banned_at_utc",
            "view_count",
            "archived",
            "no_follow",
            "is_crosspostable",
            "pinned",
            "over_18",
            "all_awardings",
            "awarders",
            "media_only",
            "can_gild",
            "spoiler",
            "locked",
            "author_flair_text",
            "treatment_tags",
            "visited",
            "removed_by",
            "num_reports",
            "distinguished",
            "author_is_blocked",
            "mod_reason_by",
            "removal_reason",
            "link_flair_background_color",
            "is_robot_indexable",
            "report_reasons",
            "author",
            "discussion_type",
            "num_comments",
            "send_replies",
            "whitelist_status",
            "contest_mode",
            "mod_reports",
            "author_patreon_flair",
            "author_flair_text_color",
            "permalink",
            "parent_whitelist_status",
            "stickied",
            "url",
            "subreddit_subscribers",
            "num_crossposts",
            "is_video",
            fields.source_keys_field,
            "collected_on",
        ]

        self.assertEqual(set(want_columns), set(got_columns), "columns match error")

        got_length = got_df.count()
        want_length = source_df.count() - 1  # filter out single bot user
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_replies_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "replies.jsonl")
        subreddit_idents = get_df(
            self.__class__.spark,
            "group_id string, id string",
            "test_data",
            "subreddit_idents.jsonl",
        )
        submission_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "submission_idents.jsonl",
        )
        reply_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "reply_idents.jsonl",
        )
        bots = get_df(
            self.__class__.spark,
            "author string",
            "test_data",
            "reddit_bots.jsonl",
        )
        got_df = from_bronze_get_replies_df(
            from_kafka_get_bronze_replies_df(source_df, "replies"),
            subreddit_idents,
            submission_idents,
            reply_idents,
            bots,
        )

        got_columns = got_df.columns
        want_columns = [
            "doc_uuid",
            "group_id",
            "submission_uuid",
            "parent_uuid",
            "created_at",
            "text",
            "all_awardings",
            "approved_at_utc",
            "approved_by",
            "archived",
            "associated_award",
            "author",
            "author_cakeday",
            "author_flair_background_color",
            "author_flair_css_class",
            "author_flair_richtext",
            "author_flair_template_id",
            "author_flair_text",
            "author_flair_text_color",
            "author_flair_type",
            "author_fullname",
            "author_is_blocked",
            "author_patreon_flair",
            "author_premium",
            "awarders",
            "banned_at_utc",
            "banned_by",
            "can_gild",
            "can_mod_post",
            "collapsed",
            "collapsed_because_crowd_control",
            "collapsed_reason",
            "collapsed_reason_code",
            "comment_type",
            "controversiality",
            "distinguished",
            "downs",
            "editable",
            "edited",
            "gilded",
            "gildings",
            "is_submitter",
            "likes",
            "link_author",
            "link_permalink",
            "link_title",
            "link_url",
            "locked",
            "mod_note",
            "mod_reason_by",
            "mod_reason_title",
            "mod_reports",
            "no_follow",
            "num_comments",
            "num_reports",
            "over_18",
            "permalink",
            "quarantine",
            "removal_reason",
            "replies",
            "report_reasons",
            "saved",
            "score",
            "score_hidden",
            "send_replies",
            "stickied",
            "subreddit_type",
            "top_awarded_type",
            "total_awards_received",
            "treatment_tags",
            "unrepliable_reason",
            "ups",
            "user_reports",
            fields.source_keys_field,
            "collected_on",
        ]

        self.assertEqual(set(want_columns), set(got_columns), "columns match error")

        got_length = got_df.count()
        want_length = source_df.count() - 4  # filter out single bot user (4 replies)
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_facebook_get_community_groups_df(self) -> None:
        group_idents = get_df(
            self.__class__.spark,
            "group_id string, id string",
            "test_data",
            "group_idents.jsonl",
        )
        got_df = from_bronze_get_community_groups_df(group_idents, "facebook")

        got_columns = got_df.columns
        want_columns = [
            fields.idents_group_id_field,
            fields.community_groups_source_id,
            fields.community_groups_source,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = group_idents.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_reddit_get_community_groups_df(self) -> None:
        subreddit_idents = get_df(
            self.__class__.spark,
            "group_id string, id string",
            "test_data",
            "subreddit_idents.jsonl",
        )
        got_df = from_bronze_get_community_groups_df(subreddit_idents, "reddit")

        got_columns = got_df.columns
        want_columns = [
            fields.idents_group_id_field,
            fields.community_groups_source_id,
            fields.community_groups_source,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = subreddit_idents.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_guilds_df_ao(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "guilds.jsonl")
        guild_idents = get_df(
            self.__class__.spark,
            "group_id string, id string",
            "test_data",
            "guild_idents.jsonl",
        )
        got_df = from_bronze_get_guilds_df_ao(
            from_kafka_get_bronze_guilds_df(source_df, "guilds"),
            guild_idents,
        )

        got_columns = got_df.columns
        want_columns = [  # copied from table bundle/src/setup_prod_catalog.ipynb
            fields.ingest_version_field,
            "group_id",
            fields.source_keys_field,
            fields.ingested_on_field,
            "name",
            "icon",
            "owner",
            "owner_id",
            "permissions",
            "features",
            "verification_level",
            "approximate_member_count",
            "approximate_presence_count",
            "description",
            "created_at",
        ]

        self.assertEqual(set(want_columns), set(got_columns), "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_channels_df_ao(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "channels.jsonl")
        got_df = from_bronze_get_channels_df_ao(
            from_kafka_get_bronze_channels_df(source_df, "channels"),
        )

        got_columns = got_df.columns
        want_columns = [  # copied from table bundle/src/setup_prod_catalog.ipynb
            fields.ingest_version_field,
            fields.source_keys_field,
            fields.ingested_on_field,
            "type",
            "position",
            "permission_overwrites",
            "name",
            "topic",
            "nsfw",
            "rate_limit_per_user",
            "icon",
            "message_count",
            "member_count",
            "thread_metadata",
            "permissions",
            "flags",
            "total_message_sent",
            "available_tags",
            "applied_tags",
            "default_reaction_emoji",
            "default_forum_layout",
            "created_at",
        ]

        self.assertEqual(set(want_columns), set(got_columns), "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_messages_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "messages.jsonl")
        message_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "message_idents.jsonl",
        )

        got_df = from_bronze_get_messages_df_ao(
            from_kafka_get_bronze_messages_df(source_df, "messages"),
            message_idents,
        )

        got_columns = got_df.columns
        want_columns = [  # copied from table bundle/src/setup_prod_catalog.ipynb
            fields.ingest_version_field,
            fields.text_version_field,
            "doc_uuid",
            fields.source_keys_field,
            fields.ingested_on_field,
            "tts",
            "mention_everyone",
            "embeds",
            "pinned",
            "type",
            "flags",
            "thread",
            "sticker_items",
            "position",
            "poll",
            "call",
            "created_at",
            "edited_at",
            "bot_content_flag",
        ]

        self.assertEqual(set(want_columns), set(got_columns), "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()

        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_discord_text_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "messages.jsonl")

        got_df = from_bronze_get_discord_text_df_ao(
            from_kafka_get_bronze_messages_df(source_df, "messages")
        )

        got_columns = got_df.columns
        want_columns = [  # copied from table bundle/src/setup_prod_catalog.ipynb
            fields.text_version_field,
            "text",
        ]

        self.assertEqual(set(want_columns), set(got_columns), "columns match error")

        got_length = got_df.count()
        want_length = source_df.count() - 29  # 29 messages with empty content

        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_discord_mention_user_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "messages.jsonl")

        got_df = from_bronze_get_discord_mention_user_df_ao(
            from_kafka_get_bronze_messages_df(source_df, "messages")
        )

        got_columns = got_df.columns
        want_columns = [  # copied from table bundle/src/setup_prod_catalog.ipynb
            fields.ingest_version_field,
            fields.source_keys_field,
            fields.ingested_on_field,
        ]

        self.assertEqual(set(want_columns), set(got_columns), "columns match error")

        got_length = got_df.count()
        want_length = 18  # 18 mentioned users

        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_discord_mention_channel_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "messages.jsonl")

        got_df = from_bronze_get_discord_mention_channel_df_ao(
            from_kafka_get_bronze_messages_df(source_df, "messages")
        )

        got_columns = got_df.columns
        want_columns = [  # copied from table bundle/src/setup_prod_catalog.ipynb
            fields.ingest_version_field,
            fields.source_keys_field,
            fields.ingested_on_field,
        ]

        self.assertEqual(set(want_columns), set(got_columns), "columns match error")

        got_length = got_df.count()
        want_length = 1  # 1 mentioned channel

        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_reactions_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "reactions.jsonl")
        got_df = from_bronze_get_reactions_df_ao(
            from_kafka_get_bronze_reactions_df(source_df, "reactions")
        )

        got_columns = got_df.columns
        want_columns = [  # copied from table bundle/src/setup_prod_catalog.ipynb
            fields.ingest_version_field,
            fields.source_keys_field,
            fields.ingested_on_field,
            "emoji",
        ]

        self.assertEqual(set(want_columns), set(got_columns), "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_members_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "members.jsonl")
        got_df = from_bronze_get_members_df_ao(
            from_kafka_get_bronze_members_df(source_df, "members")
        )

        got_columns = got_df.columns
        want_columns = [  # copied from table bundle/src/setup_prod_catalog.ipynb
            fields.ingest_version_field,
            fields.source_keys_field,
            fields.ingested_on_field,
            "nick",
            "avatar",
            "roles",
            "joined_at",
            "premium_since",
            "deaf",
            "mute",
            "pending",
            "flags",
            "permissions",
            "communication_disabled_until",
            "avatar_decoration_data",
        ]

        self.assertEqual(set(want_columns), set(got_columns), "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_discord_users_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "members.jsonl")
        got_df = from_bronze_get_users_df_ao(
            from_kafka_get_bronze_members_df(source_df, "members")
        )

        got_columns = got_df.columns
        want_columns = [  # copied from table bundle/src/setup_prod_catalog.ipynb
            fields.ingest_version_field,
            fields.source_keys_field,
            fields.ingested_on_field,
            "username",
            "global_name",
            "avatar",
            "bot",
            "system",
            "locale",
            "email",
            "flags",
        ]

        self.assertEqual(set(want_columns), set(got_columns), "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_submissions_ao_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "submissions.jsonl")
        submission_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "submission_idents.jsonl",
        )
        got_df = from_bronze_get_submissions_df_ao(
            from_kafka_get_bronze_submissions_df(source_df, "submissions"),
            submission_idents,
        )

        got_columns = got_df.columns
        want_columns = [
            "ingest_version",
            "text_version",
            "doc_uuid",
            "source_keys",
            "ingested_on",
            "source_schema",
            "approved_at_utc",
            "saved",
            "mod_reason_title",
            "gilded",
            "clicked",
            "title",
            "link_flair_richtext",
            "hidden",
            "pwls",
            "link_flair_css_class",
            "downs",
            "thumbnail_height",
            "top_awarded_type",
            "hide_score",
            "quarantine",
            "link_flair_text_color",
            "upvote_ratio",
            "author_flair_background_color",
            "subreddit_type",
            "ups",
            "total_awards_received",
            "thumbnail_width",
            "author_flair_template_id",
            "is_original_content",
            "user_reports",
            "is_reddit_media_domain",
            "is_meta",
            "category",
            "link_flair_text",
            "can_mod_post",
            "score",
            "approved_by",
            "is_created_from_ads_ui",
            "author_premium",
            "thumbnail",
            "edited",
            "author_flair_css_class",
            "author_flair_richtext",
            "gildings",
            "post_hint",
            "content_categories",
            "is_self",
            "mod_note",
            "link_flair_type",
            "wls",
            "removed_by_category",
            "banned_by",
            "author_flair_type",
            "domain",
            "allow_live_comments",
            "likes",
            "suggested_sort",
            "banned_at_utc",
            "view_count",
            "archived",
            "no_follow",
            "is_crosspostable",
            "pinned",
            "over_18",
            "all_awardings",
            "awarders",
            "media_only",
            "can_gild",
            "spoiler",
            "locked",
            "author_flair_text",
            "treatment_tags",
            "visited",
            "removed_by",
            "num_reports",
            "distinguished",
            "author_is_blocked",
            "mod_reason_by",
            "removal_reason",
            "link_flair_background_color",
            "is_robot_indexable",
            "report_reasons",
            "discussion_type",
            "num_comments",
            "send_replies",
            "whitelist_status",
            "contest_mode",
            "mod_reports",
            "author_patreon_flair",
            "author_flair_text_color",
            "permalink",
            "parent_whitelist_status",
            "stickied",
            "url",
            "subreddit_subscribers",
            "num_crossposts",
            "is_video",
            "created_at",
        ]

        self.assertEqual(set(want_columns), set(got_columns), "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_submissions_text_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "submissions.jsonl")
        got_df = from_bronze_get_submissions_text_df_ao(
            from_kafka_get_bronze_submissions_df(source_df, "submissions"),
        )

        got_columns = got_df.columns
        want_columns = [
            "text_version",
            "text",
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count() - 3  # 3 with records with null messages
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_replies_ao_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "replies.jsonl")
        reply_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "reply_idents.jsonl",
        )
        got_df = from_bronze_get_replies_df_ao(
            from_kafka_get_bronze_replies_df(source_df, "replies"),
            reply_idents,
        )

        got_columns = got_df.columns
        want_columns = [
            "ingest_version",
            "text_version",
            "doc_uuid",
            "source_keys",
            "ingested_on",
            "source_schema",
            "all_awardings",
            "approved_at_utc",
            "approved_by",
            "archived",
            "associated_award",
            "author_cakeday",
            "author_flair_background_color",
            "author_flair_css_class",
            "author_flair_richtext",
            "author_flair_template_id",
            "author_flair_text",
            "author_flair_text_color",
            "author_flair_type",
            "author_is_blocked",
            "author_patreon_flair",
            "author_premium",
            "awarders",
            "banned_at_utc",
            "banned_by",
            "can_gild",
            "can_mod_post",
            "collapsed",
            "collapsed_because_crowd_control",
            "collapsed_reason",
            "collapsed_reason_code",
            "comment_type",
            "controversiality",
            "distinguished",
            "downs",
            "editable",
            "edited",
            "gilded",
            "gildings",
            "is_submitter",
            "likes",
            "link_author",
            "link_permalink",
            "link_title",
            "link_url",
            "locked",
            "mod_note",
            "mod_reason_by",
            "mod_reason_title",
            "mod_reports",
            "no_follow",
            "num_comments",
            "num_reports",
            "over_18",
            "permalink",
            "quarantine",
            "removal_reason",
            "replies",
            "report_reasons",
            "saved",
            "score",
            "score_hidden",
            "send_replies",
            "stickied",
            "subreddit_type",
            "top_awarded_type",
            "total_awards_received",
            "treatment_tags",
            "unrepliable_reason",
            "ups",
            "user_reports",
            "created_at",
        ]

        self.assertEqual(set(want_columns), set(got_columns), "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_get_reply_text_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "replies.jsonl")
        got_df = from_bronze_get_replies_text_df_ao(
            from_kafka_get_bronze_replies_df(source_df, "replies"),
        )

        got_columns = got_df.columns
        want_columns = [
            "text_version",
            "text",
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count() - 2  # 2 with records with null messages
        self.assertEqual(want_length, got_length, "dataframe size different")
