from pyspark.sql import DataFrame

# from pyspark.sql.column import Column
import pyspark.sql.functions as fn
from pyspark.sql.types import (
    StringType,
    BinaryType,
    TimestampType,
    DoubleType,
    LongType,
)

import trend_etl.schema.fields as fields
from trend_etl.functions.misc import (
    jobid_2_timestamp,
    get_collected_day,
)
from trend_etl.functions.parsing import parse_subreddits
from trend_etl.functions.udfs import is_scientific
import trend_etl.functions.ingest_versioning as ingest
from trend_etl.schema.canonical import subreddit_canonical


# arguments:
#   archive_df: the dataframe from the archives
#   provenance: short description of the origin (kafka, s3 archive, etc.)
def from_archive_get_bronze_subreddits_df(archive_df: DataFrame) -> DataFrame:
    return archive_df.select(
        fn.col(fields.archive_value_field)
        .cast(BinaryType())
        .alias(fields.bronze_value_field),
        fn.lit("archive/reddit/subreddits").alias(fields.bronze_provenance_field),
        jobid_2_timestamp(fields.archive_job_id_field).alias(
            fields.bronze_timestamp_field
        ),
    ).withColumn(
        fields.collected_day_field,
        get_collected_day(fn.col(fields.bronze_timestamp_field)),
    )


#
# arguments:
#   input_df: dataframe received from reading a kafka topic
#   target_topic: source topic from kafka
# return: a dataframe in the schema expected for bronze tables (for all reddit data kinds)
def from_kafka_get_bronze_subreddits_df(
    input_df: DataFrame, target_topic: str
) -> DataFrame:
    return input_df.select(
        fn.col(fields.kafka_value_field).alias(fields.bronze_value_field),
        fn.lit(f"kafka/{target_topic}").alias(fields.bronze_provenance_field),
        fn.col(fields.kafka_timestamp_field).alias(fields.bronze_timestamp_field),
        get_collected_day(fn.col(fields.kafka_timestamp_field)).alias(
            fields.collected_day_field
        ),
    )


# arguments:
#   streaming_bronze: streaming dataframe from bronze table
#   subreddit_idents: groups uuid idents table
# return: dataframe in the schema of subreddits
def from_bronze_get_subreddits_df(
    streaming_bronze: DataFrame,
    subreddit_idents: DataFrame,
) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_subreddits(fn.col(fields.bronze_value_field).cast(StringType())).alias(
            "value"
        ),
        fn.col(fields.bronze_timestamp_field),
    )
    # NOTE: This filter is necessary because the bronze table
    #       for subreddits got kind=Listing documents appended.
    #       For more info see the below bug report:
    #       https://trend-community.monday.com/boards/3371582960/pulses/7142820924
    filtered = parsed_df.where(fn.col("value.kind") == fn.lit("t5"))
    flattened = filtered.select(
        fn.col("value.data.*"),
        fn.col(fields.bronze_timestamp_field),
    )

    # NOTE: left outer join (below) will cause any nulls to be
    #       detected as a downstream error.

    with_subreddit_idents = (
        flattened.join(
            subreddit_idents.withColumnRenamed(fields.idents_source_id_field, "name"),
            "name",
            "left_outer",
        )
        .withColumn(fields.source_keys_field, fn.struct(fn.col("name").alias("name")))
        .drop("name", "id")
    )

    subreddits_df = with_subreddit_idents.withColumn(
        fields.reddit_submissions_created_at,
        fn.when(
            is_scientific(fn.col("created_utc")),
            fn.col("created_utc").cast(DoubleType()).cast(LongType()),
        )
        .otherwise(fn.col("created_utc").cast(LongType()))
        .cast(TimestampType()),
    ).drop("created_utc")

    return subreddits_df


def from_bronze_get_subreddits_df_ao(
    streaming_bronze: DataFrame,
    subreddit_idents: DataFrame,
) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_subreddits(fn.col(fields.bronze_value_field).cast(StringType())).alias(
            "json"
        ),
        fn.col(fields.bronze_timestamp_field).alias(fields.ingested_on_field),
    )
    # NOTE: This filter is necessary because the bronze table
    #       for subreddits got kind=Listing documents appended.
    #       For more info see the below bug report:
    #       https://trend-community.monday.com/boards/3371582960/pulses/7142820924
    filtered = parsed_df.where(fn.col("json.kind") == fn.lit("t5"))

    flattened = filtered.select(
        ingest.ingest_version(
            fn.col("json").getField("data"), subreddit_canonical
        ).alias(fields.ingest_version_field),
        fn.col("json.data.*"),
        fn.col(fields.ingested_on_field),
    )

    # NOTE: left outer join (below) will cause any nulls to be
    #       detected as a downstream error.

    subreddits_df = (
        flattened.join(
            subreddit_idents.withColumnRenamed(fields.idents_source_id_field, "name"),
            "name",
            "left_outer",
        )
        .withColumn(fields.source_keys_field, fn.struct(fn.col("name").alias("name")))
        .withColumn(
            fields.reddit_submissions_created_at,
            fn.when(
                is_scientific(fn.col("created_utc")),
                fn.col("created_utc").cast(DoubleType()).cast(LongType()),
            )
            .otherwise(fn.col("created_utc").cast(LongType()))
            .cast(TimestampType()),
        )
        .drop("name", "id", "created_utc")
    )

    return subreddits_df
